"""
Different functions for finding the connected components of a voxel structure.
First, a sparse undirected graph is constructed from the voxel structure.
Afterward, the connected components of the graph are extracted.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import scipy.sparse as sps
import scipy.sparse.csgraph as csg


def _get_all_indices(domain_def):
    """
    Get the indices of the non-empty voxels for all the domains.
    """

    # init
    idx = np.empty(0, dtype=np.int_)

    # get the indices and colors
    for tag in domain_def:
        idx_tmp = domain_def[tag]
        idx = np.append(idx, idx_tmp)

    return idx


def _get_tag_indices(domain_def, domain_list):
    """
    Get the indices of the non-empty voxels for the given domains.
    """

    # init
    idx = np.empty(0, dtype=np.int_)

    # get the indices and colors
    for tag in domain_list:
        idx_tmp = domain_def[tag]
        idx = np.append(idx, idx_tmp)

    return idx


def _get_connection_matrix(n):
    """
    Get a sparse matrix describing the connection between the voxels.
    This matrix is extracted for the complete voxel structure.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    nv = nx*ny*nz

    # voxel index array
    x = np.arange(nx, dtype=np.int_)
    y = np.arange(ny, dtype=np.int_)
    z = np.arange(nz, dtype=np.int_)
    (idx_x, idx_y, idx_z) = np.meshgrid(x, y, z, indexing="ij")

    # voxel index number
    idx = idx_x+idx_y*nx+idx_z*nx*ny

    # create the sparse matrix
    A_connection = sps.csc_matrix((nv, nv), dtype=np.int_)

    # connections along x direction
    idx_col = idx[0:-1, :, :].flatten()
    idx_row = idx[1:, :, :].flatten()
    data = np.ones((nx-1)*ny*nz, dtype=np.int_)
    A_connection += sps.csc_matrix((data, (idx_row, idx_col)), shape=(nv, nv), dtype=np.int_)

    # connections along y direction
    idx_col = idx[:, 0:-1, :].flatten()
    idx_row = idx[:, 1:, :].flatten()
    data = np.ones(nx*(ny-1)*nz, dtype=np.int_)
    A_connection += sps.csc_matrix((data, (idx_row, idx_col)), shape=(nv, nv), dtype=np.int_)

    # connections along z direction
    idx_col = idx[:, :, 0:-1].flatten()
    idx_row = idx[:, :, 1:].flatten()
    data = np.ones(nx*ny*(nz-1), dtype=np.int_)
    A_connection += sps.csc_matrix((data, (idx_row, idx_col)), shape=(nv, nv), dtype=np.int_)

    return A_connection


def _check_domain_connection(domain_def, connection_def, tag_connection, domain_connection):
    """
    Check that the given connections between the domain exists.
    """

    # extract the data
    domain_group = domain_connection["domain_group"]
    connected = domain_connection["connected"]

    # remove empty domains
    idx_group = []
    for domain_group_tmp in domain_group:
        idx_tmp = _get_tag_indices(domain_def, domain_group_tmp)
        idx_group.append(idx_tmp)

    # init the connection matrix
    matrix = np.full((len(connection_def), len(idx_group)), True, dtype=bool)

    # fill the connection matrix
    for i, idx_graph in enumerate(connection_def):
        for j, idx_domain in enumerate(idx_group):
            idx_shared = np.intersect1d(idx_graph, idx_domain)
            matrix[i, j] = len(idx_shared) > 0

    # count connection
    vector = np.count_nonzero(matrix, axis=1)

    # check validity
    if connected:
        idx_ok = vector == len(idx_group)
        if not np.any(idx_ok):
            raise RuntimeError("domain connection: domain connection is missing: %s" % tag_connection)
    else:
        idx_ok = np.logical_or(vector == 0, vector == 1)
        if not np.all(idx_ok):
            raise RuntimeError("domain connection: domain connection is illegal: %s" % tag_connection)


def get_connection(n, domain_def, domain_connection):
    """
    Find the connected components of a voxel structure.
    """

    # get the indices of the non-empty voxels
    idx = _get_all_indices(domain_def)

    # get the connection matrix between the voxels
    A_connection = _get_connection_matrix(n)

    # get the graph matrix
    A_graph = A_connection
    A_graph = A_graph[idx, :]
    A_graph = A_graph[:, idx]

    # find the connected components in the graph
    (n_comp, labels) = csg.connected_components(
        csgraph=A_graph,
        directed=False,
        return_labels=True,
    )

    # get the indices of the connected components
    connection_def = []
    for i in range(n_comp):
        idx_local = labels == i
        idx_graph = idx[idx_local]
        connection_def.append(idx_graph)

    # check the connections between the domains
    for tag, domain_connection_tmp in domain_connection.items():
        _check_domain_connection(domain_def, connection_def, tag, domain_connection_tmp)

    return connection_def



"""
Module for solving a sparse equation system with GMRES or GCROT.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import scipy.sparse.linalg as sla


def get_solve(sol_init, op_sys, op_pcd, rhs, fct_callback, iter_options):
    """
    Solve a sparse equation system with GMRES or GCROT (main function).
    The equation system and the preconditioner are described with linear operator.
    """

    # get the options
    solver = iter_options["solver"]
    rel_tol = iter_options["rel_tol"]
    abs_tol = iter_options["abs_tol"]
    n_inner = iter_options["n_inner"]
    n_outer = iter_options["n_outer"]

    # call the solver
    if solver == "gmres":
        (sol, flag) = sla.gmres(
            op_sys, rhs, M=op_pcd, x0=sol_init,
            tol=rel_tol, atol=abs_tol,
            restart=n_inner, maxiter=n_outer,
            callback=fct_callback, callback_type="x",
        )
    elif solver == "gcrot":
        (sol, flag) = sla.gcrotmk(
            op_sys, rhs, M=op_pcd, x0=sol_init,
            tol=rel_tol, atol=abs_tol,
            m=n_inner, maxiter=n_outer,
            callback=fct_callback,
        )
    else:
        raise ValueError("invalid matrix solver")

    # check for convergence
    status = flag == 0

    return status, sol



"""
Module for doing matrix-vector multiplication (with circulant tensors and FFT).

The multiplication can be done in two ways:
    - combined: the computation is done with the 4D tensors
    - split: the 4D tensors are sliced into several 3D tensors

The "combined" mode is typically faster than the "split" mode.
The "split" mode features a reduced memory footprint.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_matrix import fourier_transform
from pypeec import log

# get a logger
LOGGER = log.get_logger("FFT")

# dummy options
LIBRARY = None
MATRIX_SPLIT = None
USE_FFT_GPU = None
NPCP = None


def _get_tensor_sign(name, nd_in):
    """
    Get the signs for the different tensor blocks composing the circulant tensor.
    """

    if name == "potential":
        sign = NPCP.ones((2, 2, 2, nd_in), dtype=NPCP.float_)
    elif name == "inductance":
        sign = NPCP.ones((2, 2, 2, nd_in), dtype=NPCP.float_)
    elif name == "coupling":
        sign = NPCP.empty((2, 2, 2, nd_in), dtype=NPCP.float_)
        sign[0, 0, 0, :] = NPCP.array([+1, +1, +1], dtype=NPCP.float_)
        sign[1, 0, 0, :] = NPCP.array([-1, +1, +1], dtype=NPCP.float_)
        sign[0, 1, 0, :] = NPCP.array([+1, -1, +1], dtype=NPCP.float_)
        sign[0, 0, 1, :] = NPCP.array([+1, +1, -1], dtype=NPCP.float_)
        sign[1, 1, 0, :] = NPCP.array([-1, -1, +1], dtype=NPCP.float_)
        sign[1, 0, 1, :] = NPCP.array([-1, +1, -1], dtype=NPCP.float_)
        sign[0, 1, 1, :] = NPCP.array([+1, -1, -1], dtype=NPCP.float_)
        sign[1, 1, 1, :] = NPCP.array([-1, -1, -1], dtype=NPCP.float_)
    else:
        raise ValueError("invalid matrix type")

    return sign


def _get_tensor_circulant(mat, sign):
    """
    Construct a circulant tensor from a 4D tensor.
    The circulant tensor is constructed for the first 3D.

    The input tensor has the size: (nx, ny, nz, nd_in).
    The output FFT circulant tensor has the size: (2*nx, 2*ny, 2*nz, nd_in).
    """

    # get the tensor size
    (nx, ny, nz, nd_in) = mat.shape

    # init the circulant tensor
    mat_fft = NPCP.zeros((2*nx, 2*ny, 2*nz, nd_in), dtype=NPCP.float_)

    # cube none
    mat_fft[0:nx, 0:ny, 0:nz, :] = mat[0:nx, 0:ny, 0:nz, :]*sign[0:1, 0:1, 0:1, :]
    # cube x
    mat_fft[nx+1:2*nx, 0:ny, 0:nz, :] = mat[nx-1:0:-1, 0:ny, 0:nz, :]*sign[1:2, 0:1, 0:1, :]
    # cube y
    mat_fft[0:nx, ny+1:2*ny, 0:nz, :] = mat[0:nx, ny-1:0:-1, 0:nz, :]*sign[0:1, 1:2, 0:1, :]
    # cube z
    mat_fft[0:nx, 0:ny, nz+1:2*nz, :] = mat[0:nx, 0:ny, nz-1:0:-1, :]*sign[0:1, 0:1, 1:2, :]
    # cube xy
    mat_fft[nx+1:2*nx, ny+1:2*ny, 0:nz, :] = mat[nx-1:0:-1, ny-1:0:-1, 0:nz, :]*sign[1:2, 1:2, 0:1, :]
    # cube xz
    mat_fft[nx+1:2*nx, 0:ny, nz+1:2*nz, :] = mat[nx-1:0:-1, 0:ny, nz-1:0:-1, :]*sign[1:2, 0:1, 1:2, :]
    # cube yz
    mat_fft[0:nx, ny+1:2*ny, nz+1:2*nz, :] = mat[0:nx, ny-1:0:-1, nz-1:0:-1, :]*sign[0:1, 1:2, 1:2, :]
    # cube xyz
    mat_fft[nx+1:2*nx, ny+1:2*ny, nz+1:2*nz, :] = mat[nx-1:0:-1, ny-1:0:-1, nz-1:0:-1, :]*sign[1:2, 1:2, 1:2, :]

    # get the FFT of the circulant tensor
    mat_fft = fourier_transform.get_fft_tensor_keep(mat_fft, True)

    return mat_fft


def _get_indices(nx, ny, nz, idx, nd_out, dim):
    """
    Get the indices for mapping a vector into a tensor.
    The indices are either computer for all 4D or for a 3D slice.
    """

    if dim is None:
        # shape of the tensor with the vectors (4D)
        shape = (nx, ny, nz, nd_out)

        # mapping between the vector indices and the tensor indices (4D)
        idx_mat = NPCP.unravel_index(idx, (nx, ny, nz, nd_out), order="F")

        # for the case with 4D mapping, all the elements are selected
        idx_sel = None
    else:
        # number of element for the 3D sluce
        nv = nx*ny*nz

        # shape of the tensor with the vectors (4D)
        shape = (nx, ny, nz)

        # indices of the elements included in the considered 3D slices
        idx_sel = NPCP.in1d(idx, NPCP.arange(dim*nv, (dim+1)*nv))

        # mapping between the vector indices and the tensor indices (3D)
        idx_tmp = idx[idx_sel]-dim*nv
        idx_mat = NPCP.unravel_index(idx_tmp, shape, order="F")

    # assign the dict with the indices
    idx = {"idx_sel": idx_sel, "idx_mat": idx_mat, "shape": shape, "length": len(idx)}

    return idx


def _get_tensor(idx, vec):
    """
    Transform a vector into a tensor.
    This is used for the input vector.
    """

    # extract the mapping data
    shape = idx["shape"]
    idx_sel = idx["idx_sel"]
    idx_mat = idx["idx_mat"]

    # init the tensor
    res = NPCP.zeros(shape, dtype=NPCP.complex_)

    # assign the tensor (4D or 3D slice)
    if idx_sel is None:
        res[idx_mat] = vec[:]
    else:
        res[idx_mat] = vec[idx_sel]

    return res


def _get_vector(idx, res):
    """
    Transform a tensor into a vector.
    This is used for the output vector.
    """

    # extract the mapping data
    length = idx["length"]
    idx_sel = idx["idx_sel"]
    idx_mat = idx["idx_mat"]

    # init the vector
    vec = NPCP.zeros(length, dtype=NPCP.complex_)

    # assign the vector (4D or 3D slice)
    if idx_sel is None:
        vec[:] = res[idx_mat]
    else:
        vec[idx_sel] = res[idx_mat]

    return vec


def _get_compute_combined(name, idx_in, idx_out, mat_fft, vec_in):
    """
    Matrix-vector multiplication with FFT.
    The multiplication is done directly with the 4D tensors.

    The output index vector has the size: n_out.
    The input index vector has the size: n_in.
    The input vector has the size: n_in.
    The output vector has the size: n_out.

    The FFT circulant tensor has the size: (2*nx, 2*ny, 2*nz, nd_in).
    The input tensor has the size: (nx, ny, nz, nd_out).

    For the matrix-vector multiplication is done in several steps:
        - the vector is expanded into a tensor: n_in to (nx, ny, nz, nd_out)
        - computation the FFT of the obtained tensor: (nx, ny, nz, nd_out) to (2*nx, 2*ny, 2*nz, nd_out)
        - multiplication of FFT circulant tensors: (2*nx, 2*ny, 2*nz, nd_in) and (2*nx, 2*ny, 2*nz, nd_out)
        - computation the iFFT of the obtained tensor: (2*nx, 2*ny, 2*nz, nd_out)
        - the tensor is flattened into a vector: (2*nx, 2*ny, 2*nz, nd_out) to n_out
    """

    # get the input tensor from the input vector
    res = _get_tensor(idx_in, vec_in)

    # compute the FFT of the input tensor
    res = fourier_transform.get_fft_tensor_expand(res, True)

    # matrix vector multiplication in frequency domain with the FFT circulant tensor
    if name == "potential":
        res *= mat_fft
    elif name == "inductance":
        res *= mat_fft
    elif name == "coupling":
        res_tmp = NPCP.empty(res.shape, dtype=NPCP.complex_)
        res_tmp[:, :, :, 0] = +mat_fft[:, :, :, 2]*res[:, :, :, 1]+mat_fft[:, :, :, 1]*res[:, :, :, 2]
        res_tmp[:, :, :, 1] = -mat_fft[:, :, :, 2]*res[:, :, :, 0]+mat_fft[:, :, :, 0]*res[:, :, :, 2]
        res_tmp[:, :, :, 2] = -mat_fft[:, :, :, 1]*res[:, :, :, 0]-mat_fft[:, :, :, 0]*res[:, :, :, 1]
        res = res_tmp
    else:
        raise ValueError("invalid matrix type")

    # compute the iFFT of the obtained output tensor
    res = fourier_transform.get_ifft_tensor(res, True)

    # extract the output vector from the output tensor
    res = _get_vector(idx_out, res)

    return res


def _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, dim_in, dim_out, dim_mat):
    """
    Matrix-vector multiplication with FFT.
    The multiplication is done for specific 3D slices composing the 4D tensors.
    """

    # get the input tensor from the input vector
    res = _get_tensor(idx_in[dim_in], vec_in)

    # compute the FFT of the input tensor
    res = fourier_transform.get_fft_tensor_expand(res, True)

    # matrix vector multiplication in frequency domain with the FFT circulant tensor
    res *= mat_fft[:, :, :, dim_mat]

    # compute the iFFT of the obtained output tensor
    res = fourier_transform.get_ifft_tensor(res, True)

    # extract the output vector from the output tensor
    res = _get_vector(idx_out[dim_out], res)

    return res


def _get_compute_split(name, n_out, idx_in, idx_out, mat_fft, vec_in):
    """
    Matrix-vector multiplication with FFT.
    The multiplication is done by splitting the 4D tensor in 3D slices.

    The output index vector has the size: n_out.
    The input index vector has the size: n_in.
    The input vector has the size: n_in.
    The output vector has the size: n_out.

    The FFT circulant tensor has the size: (2*nx, 2*ny, 2*nz, nd_in).
    The input tensor has the size: (nx, ny, nz, nd_out).
    The dimension nd_in and nd_out are used to create the 3D slices.

    For the matrix-vector multiplication is done in several steps for each slice:
        - the vector is expanded into a tensor: n_in to (nx, ny, nz)
        - computation the FFT of the obtained tensor: (nx, ny, nz) to (2*nx, 2*ny, 2*nz)
        - multiplication of FFT circulant tensors: (2*nx, 2*ny, 2*nz) and (2*nx, 2*ny, 2*nz)
        - computation the iFFT of the obtained tensor: (2*nx, 2*ny, 2*nz)
        - the tensor is flattened into a vector: (2*nx, 2*ny, 2*nz) to n_out
    """

    if name == "potential":
        # the multiplication is composed of a single slice
        res = _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 0, 0, 0)
    elif name == "inductance":
        # the multiplication is decomposed into three slices
        res = NPCP.zeros(n_out, dtype=NPCP.complex_)
        res += _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 0, 0, 0)
        res += _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 1, 1, 0)
        res += _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 2, 2, 0)
    elif name == "coupling":
        # the multiplication is decomposed into six slices
        res = NPCP.zeros(n_out, dtype=NPCP.complex_)
        res += _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 1, 0, 2)
        res += _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 2, 0, 1)
        res += _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 2, 1, 0)
        res -= _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 0, 1, 2)
        res -= _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 0, 2, 1)
        res -= _get_multiply_slice(idx_in, idx_out, mat_fft, vec_in, 1, 2, 0)
    else:
        raise ValueError("invalid matrix type")

    return res


def _get_prepare_sub(name, idx_out, idx_in, mat):
    """
    Construct a circulant tensor from a 4D tensor (main function).
    The circulant tensor is constructed along the first 3D.
    The indices for mapping a vector into a tensor are computed.

    The input tensor has the size: (nx, ny, nz, nd_in).
    The output FFT circulant tensor has the size: (2*nx, 2*ny, 2*nz, nd_in).
    """

    # get tensor size
    (nx, ny, nz, nd_in) = mat.shape

    # get the memory footprint
    nnz = (2*nx)*(2*ny)*(2*nz)*nd_in
    itemsize = NPCP.dtype(NPCP.complex_).itemsize
    footprint = (itemsize*nnz)/(1024**2)

    # display the tensor size
    LOGGER.debug("tensor size: (%d, %d, %d)" % (nx, ny, nz))
    LOGGER.debug("tensor footprint: %.2f MB" % footprint)

    # get the sign that will be applied to the different blocks of the tensor
    sign = _get_tensor_sign(name, nd_in)

    # load the data to the GPU
    if USE_FFT_GPU:
        mat = NPCP.asarray(mat)
        idx_in = NPCP.asarray(idx_in)
        idx_out = NPCP.asarray(idx_out)

    # get the FFT circulant tensor
    mat_fft = _get_tensor_circulant(mat, sign)

    # get tensor last dimension
    if name == "potential":
        nd_out = 1
    elif name == "inductance":
        nd_out = 3
    elif name == "coupling":
        nd_out = 3
    else:
        raise ValueError("invalid matrix type")

    # length of the output
    n_in = len(idx_in)
    n_out = len(idx_out)

    # compute the indices for mapping a vector into a tensor
    if MATRIX_SPLIT:
        # the following method is used for the multiplication
        #   - 4D tensor will be sliced into 3D tensors for the computation
        #   - compute the indices for each 3D slice
        idx_in_mat = []
        idx_out_mat = []
        for i in range(nd_out):
            idx_in_mat.append(_get_indices(nx, ny, nz, idx_in, nd_out, i))
            idx_out_mat.append(_get_indices(nx, ny, nz, idx_out, nd_out, i))
    else:
        # the following method is used for the multiplication
        #   - 4D tensor are directly used for the computation
        #   - compute the indices for the 4D tensor
        idx_in_mat = _get_indices(nx, ny, nz, idx_in, nd_out, None)
        idx_out_mat = _get_indices(nx, ny, nz, idx_out, nd_out, None)

    return name, n_in, n_out, idx_in_mat, idx_out_mat, mat_fft


def get_prepare(name, idx_out, idx_in, mat):
    """
    Construct a circulant tensor from a 4D tensor (log wrapper).
    """

    LOGGER.debug("multiplication: %s" % name)
    with log.BlockIndent():
        data = _get_prepare_sub(name, idx_out, idx_in, mat)

    return data


def set_options(fft_options):
    """
    Assign the options and load the right libray.
    """

    # assign global variable
    global LIBRARY
    global MATRIX_SPLIT
    LIBRARY = fft_options["library"]
    MATRIX_SPLIT = fft_options["matrix_split"]

    # import the right library
    global NPCP
    global USE_FFT_GPU

    if LIBRARY == "CuPy":
        import cupy as lib_tmp
        NPCP = lib_tmp
        USE_FFT_GPU = True
    elif LIBRARY in ["SciPy", "FFTW"]:
        import numpy as lib_tmp
        NPCP = lib_tmp
        USE_FFT_GPU = False
    else:
        raise ValueError("invalid FFT library")


def get_multiply(data, vec_in, flip):
    """
    Matrix-vector multiplication with FFT.
    If the flip switch is activated, the input and output are flipped.

    The output index vector has the size: n_out.
    The input index vector has the size: n_in.
    The input vector has the size: n_in.
    The output vector has the size: n_out.
    """

    # extract the data
    (name, n_in, n_out, idx_in, idx_out, mat_fft) = data

    # flip the input and output
    if flip:
        (n_out, n_in) = (n_in, n_out)
        (idx_out, idx_in) = (idx_in, idx_out)

    # load the data to the GPU
    if USE_FFT_GPU:
        vec_in = NPCP.array(vec_in)

    if MATRIX_SPLIT:
        vec_out = _get_compute_split(name, n_out, idx_in, idx_out, mat_fft, vec_in)
    else:
        vec_out = _get_compute_combined(name, idx_in, idx_out, mat_fft, vec_in)

    # unload the data from the GPU
    if USE_FFT_GPU:
        vec_out = NPCP.asnumpy(vec_out)

    return vec_out



"""
Script for plotting the geometry of the examples.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os.path
from pypeec import main
from pypeec import io

# get the path of the root of the code
PATH_ROOT = os.path.dirname(__file__)


def run_image(folder_example):
    """
    Run the mesher for a specified example.
    Plot the geometry with an empty background.
    """

    # example folder
    folder_base = os.path.join(PATH_ROOT, "..", "..", "examples")

    # get the geometry file
    file_geometry = os.path.join(folder_base, folder_example, "geometry.yaml")

    # get the viewer file
    file_viewer = os.path.join(folder_base, "config", "viewer.yaml")

    # load data
    data_viewer = io.load_input(file_viewer)
    data_geometry = io.load_input(file_geometry)

    # tweak the plot options
    data_viewer["domain"]["data_window"]["show_menu"] = True
    data_viewer["domain"]["data_window"]["window_size"] = [400, 300]
    data_viewer["domain"]["data_options"]["plot_theme"]["background_color"] = "white"
    data_viewer["domain"]["data_options"]["plot_theme"]["axis_add"] = False
    data_viewer["domain"]["data_options"]["plot_view"]["grid_plot"] = False
    data_viewer["domain"]["data_options"]["plot_view"]["geom_plot"] = False
    data_viewer["domain"]["data_options"]["plot_view"]["point_plot"] = False
    data_viewer["domain"]["data_plot"]["title"] = None

    # run the mesher
    data_voxel = main.run_mesher_data(data_geometry)

    # run the viewer
    main.run_viewer_data(
        data_voxel, data_viewer,
        plot_mode="qt", tag_plot=["domain"],
    )


if __name__ == "__main__":
    run_image("examples_png/iron_core")



"""
Module for checking the voxel mesher data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def _check_domain_index(domain_index):
    """
    Check the domain definition (mapping between domain names and indices).
    """

    # check type
    datachecker.check_dict("domain_index", domain_index, can_be_empty=False)

    # check data
    for idx in domain_index.values():
        datachecker.check_integer_array("domain_index", idx, is_positive=True, can_be_zero=True)


def _check_param(param):
    """
    Check the voxel structure parameters.
    """

    # check type
    key_list = ["n", "d", "c"]
    datachecker.check_dict("param", param, key_list=key_list)

    # check data
    datachecker.check_integer_array("n", param["n"], size=3, is_positive=True, can_be_zero=False)
    datachecker.check_float_array("d", param["d"], size=3, is_positive=True, can_be_zero=False)
    datachecker.check_float_array("c", param["c"], size=3)


def check_data_voxelize(data_voxelize):
    """
    Check the data used for voxelization.
    """

    # check type
    key_list = ["param", "domain_index"]
    datachecker.check_dict("data_voxelize", data_voxelize, key_list=key_list)

    # extract field
    param = data_voxelize["param"]
    domain_index = data_voxelize["domain_index"]

    # check voxel structure parameters
    _check_param(param)

    # check domain definition
    _check_domain_index(domain_index)

    # get the domain name
    domain_list = domain_index.keys()

    return domain_list



"""
Different functions for extracting PyVista object from the voxel structure:
    - the complete voxel structure (uniform grid)
    - the structure containing non-empty voxels (unstructured grid)
    - the defined point cloud (polydata object)
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import pyvista as pv


def get_grid(n, d, c):
    """
    Construct a PyVista uniform grid for the complete voxel structure.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    (dx, dy, dz) = d
    (cx, cy, cz) = c

    # origin coordinate
    ox = cx-(nx*dx)/2
    oy = cy-(ny*dy)/2
    oz = cz-(nz*dz)/2

    # create a uniform grid for the complete structure
    grid = pv.ImageData()

    # set the array size and the voxel size
    grid.origin = (ox, oy, oz)
    grid.dimensions = (nx+1, ny+1, nz+1)
    grid.spacing = (dx, dy, dz)

    return grid


def get_voxel(grid, idx):
    """
    Construct a PyVista unstructured grid for the non-empty voxels.
    The indices of the non-empty voxels are provided.
    """

    # assemble idx
    idx = np.sort(idx)

    # transform the uniform grid into an unstructured grid (keeping the non-empty voxels)
    voxel = grid.extract_cells(idx)

    return voxel


def get_point(pts_cloud):
    """
    Construct a PyVista point cloud (polydata) with the defined points.
    """

    point = pv.PolyData(pts_cloud)

    return point


def get_reference(reference, voxel):
    """
    Construct a PyVista object from the reference mesh.
    """

    if reference is None:
        reference = voxel
    else:
        points = reference["points"]
        faces = reference["faces"]
        reference = pv.PolyData(points, faces)

    return reference



"""
Module for checking the geometry data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker
from pypeec.lib_check import check_data_geometry_png
from pypeec.lib_check import check_data_geometry_stl
from pypeec.lib_check import check_data_geometry_shape
from pypeec.lib_check import check_data_geometry_voxel


def _check_resampling(resampling):
    """
    Check the validity of resampling parameters.
    """

    # check type
    key_list = [
        "use_reduce",
        "use_resample",
        "resampling_factor",
    ]
    datachecker.check_dict("resampling", resampling, key_list=key_list)

    # check data
    datachecker.check_boolean("use_reduce", resampling["use_reduce"])
    datachecker.check_boolean("use_resample", resampling["use_resample"])
    datachecker.check_integer_array("resampling_factor", resampling["resampling_factor"], size=3, is_positive=True, can_be_zero=False)


def _check_domain_conflict(domain_list, domain_conflict):
    """
    Check the validity of the rules to solve conflict between domains (STL mesher).
    """

    # check type
    datachecker.check_list("domain_conflict", domain_conflict, can_be_empty=True)

    # check value
    for domain_conflict_tmp in domain_conflict:
        # check type
        key_list = ["domain_keep", "domain_resolve"]
        datachecker.check_dict("domain_conflict", domain_conflict_tmp, key_list=key_list)

        # extract the data
        domain_keep = domain_conflict_tmp["domain_keep"]
        domain_resolve = domain_conflict_tmp["domain_resolve"]

        # check type
        datachecker.check_list("domain_keep", domain_keep, sub_type=str, can_be_empty=False)
        datachecker.check_list("domain_resolve", domain_resolve, sub_type=str, can_be_empty=False)

        # check data
        for tag in domain_keep:
            datachecker.check_choice("domain_keep", tag, domain_list)
        for tag in domain_resolve:
            datachecker.check_choice("domain_resolve", tag, domain_list)


def _check_domain_connection(domain_list, domain_connection):
    """
    Check the domain connection data.
    This list is defining the required connection between the domain
    """

    # check type
    datachecker.check_dict("domain_connection", domain_connection)

    # check value
    for domain_connection_tmp in domain_connection.values():
        # check type
        key_list = ["connected", "domain_group"]
        datachecker.check_dict("domain_connection", domain_connection_tmp, key_list=key_list)

        # extract field
        domain_group = domain_connection_tmp["domain_group"]
        connected = domain_connection_tmp["connected"]

        # check type
        datachecker.check_boolean("connected", connected)
        datachecker.check_list("domain_group", domain_group, can_be_empty=False)

        # check value
        for domain_group_tmp in domain_group:
            datachecker.check_list("domain_group", domain_group_tmp, sub_type=str, can_be_empty=False)
            for tag in domain_group_tmp:
                datachecker.check_choice("domain_group", tag, domain_list)


def check_data_geometry(data_geometry):
    """
    Check the mesher data type and extract the data.
    """

    # check type
    key_list = [
        "mesh_type",
        "data_voxelize",
        "resampling",
        "check_cloud",
        "check_conflict",
        "check_connection",
        "domain_conflict",
        "domain_connection",
        "pts_cloud",
    ]
    datachecker.check_dict("data_geometry", data_geometry, key_list=key_list)

    # extract field
    mesh_type = data_geometry["mesh_type"]
    data_voxelize = data_geometry["data_voxelize"]
    resampling = data_geometry["resampling"]
    check_cloud = data_geometry["check_cloud"]
    check_conflict = data_geometry["check_conflict"]
    check_connection = data_geometry["check_connection"]
    domain_conflict = data_geometry["domain_conflict"]
    domain_connection = data_geometry["domain_connection"]
    pts_cloud = data_geometry["pts_cloud"]

    # check type
    datachecker.check_choice("mesh_type", mesh_type, ["stl", "png", "shape", "voxel"])
    datachecker.check_boolean("check_cloud", check_cloud)
    datachecker.check_boolean("check_conflict", check_conflict)
    datachecker.check_boolean("check_connection", check_connection)

    # check the mesher
    if mesh_type == "png":
        domain_list = check_data_geometry_png.check_data_voxelize(data_voxelize)
    elif mesh_type == "stl":
        domain_list = check_data_geometry_stl.check_data_voxelize(data_voxelize)
    elif mesh_type == "shape":
        domain_list = check_data_geometry_shape.check_data_voxelize(data_voxelize)
    elif mesh_type == "voxel":
        domain_list = check_data_geometry_voxel.check_data_voxelize(data_voxelize)
    else:
        raise ValueError("invalid mesh type")

    # check the resampling data
    _check_resampling(resampling)

    # check the conflict data
    _check_domain_conflict(domain_list, domain_conflict)

    # check the connection data
    _check_domain_connection(domain_list, domain_connection)

    # check the point cloud
    datachecker.check_float_pts("pts_cloud", pts_cloud, size=3, can_be_empty=True)



"""
Module for handling voxel indices (linear and tensor indices).
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np


def get_idx_linear(n, idx_tensor):
    """
    Transform tensor indices into linear indices.
    """

    # extract the voxel data
    (nx, ny, nz) = n

    # extract indices
    idx_x = idx_tensor[:, 0]
    idx_y = idx_tensor[:, 1]
    idx_z = idx_tensor[:, 2]

    # check bounds
    if not (np.all(idx_x >= 0) and np.all(idx_x < nx)):
        raise ValueError("invalid index range (x coordinate)")
    if not (np.all(idx_y >= 0) and np.all(idx_y < nx)):
        raise ValueError("invalid index range (y coordinate)")
    if not (np.all(idx_z >= 0) and np.all(idx_z < nx)):
        raise ValueError("invalid index range (z coordinate)")

    # convert tensor indices into linear indices
    idx_linear = idx_x+nx*idx_y+nx*ny*idx_z

    return idx_linear


def get_idx_tensor(n, idx_linear):
    """
    Transform linear indices into tensor indices.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    nv = nx*ny*nz

    # check bounds
    if not (np.all(idx_linear >= 0) and np.all(idx_linear < nv)):
        raise ValueError("invalid index range")

    # convert linear indices into tensor indices
    (idx_x, idx_y, idx_z) = np.unravel_index(idx_linear, (nx, ny, nz), order="F")

    # assemble the coordinate array
    idx_tensor = np.stack((idx_x, idx_y, idx_z), axis=1)

    return idx_tensor



"""
Module for handling the logging.
    - Use a global timer to measure the elapsed time.
    - Provide a class for timing (and logging) code blocks.
    - Measure duration with local timers.
    - Log exceptions.

The log config is defined by the following files.
    - First the default configuration is loaded ("pypeec/data/logger.yaml").
    - Afterward, a custom file can be loaded with an environment variable ("PYTHONLOGGER").
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os
import sys
import datetime
import threading
import logging
import yaml
import importlib.resources

# global timestamp (constant over the complete run)
GLOBAL_TIMESTAMP = datetime.datetime.today()

# logging indentation level (updated inside the blocks)
GLOBAL_LEVEL = 0


def _check_boolean(name, data):
    """
    Check a boolean.
    """

    if not isinstance(data, bool):
        raise AssertionError("%s: should be a boolean" % name)


def _check_integer(name, data):
    """
    Check a integer.
    """

    # check type
    if not isinstance(data, int):
        raise AssertionError("%s: should be an integer" % name)

    # check value
    if data < 0:
        raise AssertionError("%s: should be positive" % name)


def _check_string(name, data,):
    """
    Check a string.
    """

    # check type
    if not isinstance(data, str):
        raise AssertionError("%s: should be a string" % name)
    if len(data) == 0:
        raise AssertionError("%s: cannot be empty" % name)


def _check_dict(name, data, key_list):
    """
    Check a dict.
    """

    # check type
    if not isinstance(data, dict):
        raise AssertionError("%s: should be a dict" % name)

    # check keys
    for tag in key_list:
        if tag not in data:
            raise AssertionError("%s: dict is incomplete: %s" % (name, tag))


def _check_config(data):
    """
    Check the integrity of the config file.
    """

    # check dict
    key_list = [
        "FORMAT",
        "LEVEL",
        "INDENTATION",
        "EXCEPTION_TRACE",
        "USE_COLOR",
        "DEF_COLOR",
    ]
    _check_dict("data", data, key_list)
    
    # check data
    _check_string("LEVEL", data["LEVEL"])
    _check_integer("INDENTATION", data["INDENTATION"])
    _check_boolean("EXCEPTION_TRACE", data["EXCEPTION_TRACE"])
    _check_boolean("USE_COLOR", data["USE_COLOR"])

    # check sub dict
    key_list = [
        "LOGGER",
        "TIMESTAMP_FMT",
        "DURATION_FMT",
        "TIMESTAMP_TRC",
        "DURATION_TRC",
    ]
    _check_dict("FORMAT", data["FORMAT"], key_list)
    _check_string("LOGGER", data["FORMAT"]["LOGGER"])
    _check_string("TIMESTAMP_FMT", data["FORMAT"]["TIMESTAMP_FMT"])
    _check_string("DURATION_FMT", data["FORMAT"]["DURATION_FMT"])
    _check_integer("TIMESTAMP_TRC", data["FORMAT"]["TIMESTAMP_TRC"])
    _check_integer("DURATION_TRC", data["FORMAT"]["DURATION_TRC"])

    # check sub dict
    key_list = [
        "CL_DEBUG",
        "CL_INFO",
        "CL_WARNING",
        "CL_ERROR",
        "CL_CRITICAL",
        "CL_RESET",
    ]
    _check_dict("DEF_COLOR", data["DEF_COLOR"], key_list)
    _check_string("CL_DEBUG", data["DEF_COLOR"]["CL_DEBUG"])
    _check_string("CL_INFO", data["DEF_COLOR"]["CL_INFO"])
    _check_string("CL_WARNING", data["DEF_COLOR"]["CL_WARNING"])
    _check_string("CL_ERROR", data["DEF_COLOR"]["CL_ERROR"])
    _check_string("CL_CRITICAL", data["DEF_COLOR"]["CL_CRITICAL"])
    _check_string("CL_RESET", data["DEF_COLOR"]["CL_RESET"])


def _get_fmt(color, reset):
    """
    Get a logging formatter.
    """

    if (color is None) and (reset is None):
        fmt = logging.Formatter(FORMAT["LOGGER"])
    else:
        fmt = logging.Formatter("\x1b" + color + FORMAT["LOGGER"] + "\x1b" + reset)

    return fmt


def _get_format_timestamp(timestamp):
    """
    Format a timestamp into a string.
    """

    timestamp_str = timestamp.strftime(FORMAT["TIMESTAMP_FMT"])
    timestamp_str = timestamp_str[0:len(timestamp_str)-FORMAT["TIMESTAMP_TRC"]]

    return timestamp_str


def _get_format_duration(duration):
    """
    Format the duration into a string.
    """

    # get reference timestamp
    timestamp = datetime.datetime.utcfromtimestamp(0)

    # convert timestamp to duration
    timestamp = timestamp+duration

    # format timestamp
    duration_str = timestamp.strftime(FORMAT["DURATION_FMT"])
    duration_str = duration_str[0:len(duration_str)-FORMAT["DURATION_TRC"]]

    return duration_str


class _DeltaTimeFormatter(logging.Formatter):
    """
    Class for adding elapsed time to a logger.
    """

    def __init__(self):
        """
        Constructor.
        Create a timer.
        """

        # call parent constructor
        super().__init__()

        # define the color formatters
        self.fmt_color = {
            logging.DEBUG: _get_fmt(DEF_COLOR["CL_DEBUG"], DEF_COLOR["CL_RESET"]),
            logging.INFO: _get_fmt(DEF_COLOR["CL_INFO"], DEF_COLOR["CL_RESET"]),
            logging.WARNING: _get_fmt(DEF_COLOR["CL_WARNING"], DEF_COLOR["CL_RESET"]),
            logging.ERROR: _get_fmt(DEF_COLOR["CL_ERROR"], DEF_COLOR["CL_RESET"]),
            logging.CRITICAL: _get_fmt(DEF_COLOR["CL_CRITICAL"], DEF_COLOR["CL_RESET"]),
        }

        # define the black formatter
        self.fmt_black = _get_fmt(None, None)

    def format(self, record):
        """
        Format a record to a string.
        Add the elapsed time.
        """

        # get log
        lvl = record.levelno
        msg = record.msg

        # add the elapsed time to the log record
        timestamp = datetime.datetime.today()
        duration = timestamp-GLOBAL_TIMESTAMP
        record.timestamp = _get_format_timestamp(timestamp)
        record.duration = _get_format_duration(duration)

        # add the process and thread id to the log record
        record.process_id = os.getpid()
        record.thread_id = threading.get_native_id()

        # get the message padding for the desired indentation
        pad = " " * (GLOBAL_LEVEL*INDENTATION)

        # add the padding to the message
        record.msg = pad + msg

        # get the formatter
        if USE_COLOR:
            msg = self.fmt_color[lvl].format(record)
        else:
            msg = self.fmt_black.format(record)

        return msg


class BlockTimer:
    """
    Class for timing a block of code.
        - Uses enter and exit magic methods.
        - Display the results with a logger.

    Parameters
    ----------
    logger : logger
        Logger object instance.
    name : string
        Name of the code block.
    level : string
        Logging level to be used.
    """

    def __init__(self, logger, name, level="INFO"):
        """
        Constructor.
        Set the logger.
        """

        self.logger = logger
        self.name = name
        self.level = logging.getLevelName(level)
        self.timestamp = None

    def __enter__(self):
        """
        Enter magic method.
        Reset the timer and log the results.
        """

        # start the timer and display
        self.timestamp = datetime.datetime.today()
        self.logger.log(self.level, self.name + " : enter : timing")

        # increase the indentation of the block
        global GLOBAL_LEVEL
        GLOBAL_LEVEL += 1

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Exit magic method.
        Get the elapsed time and log the results.
        """

        # restore the indentation to the previous state
        global GLOBAL_LEVEL
        GLOBAL_LEVEL -= 1

        # get timing
        duration = datetime.datetime.today()-self.timestamp
        duration = _get_format_duration(duration)

        # display exit message
        self.logger.log(self.level, self.name + " : exit : " + duration)


class BlockIndent:
    """
    Class for indenting a block of code.
        - Uses enter and exit magic methods.
        - Display the results with a logger.
    """

    def __init__(self):
        """
        Constructor.
        """

        pass

    def __enter__(self):
        """
        Enter magic method.
        Increase the indentation of the block.
        """

        global GLOBAL_LEVEL
        GLOBAL_LEVEL += 1

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Exit magic method.
        Restore the indentation to the previous state.
        """

        global GLOBAL_LEVEL
        GLOBAL_LEVEL -= 1


def log_exception(logger, ex, level="ERROR"):
    """
    Log an exception.
        - Log the exception type, message, and trace.
        - Remove the context from the exception before the logging.

    Parameters
    ----------
    logger : logger
        Logger object instance.
    ex : exception
        Exception to be logged.
    level : string
        Logging level to be used.
    """

    # remove the expression context
    ex.__context__ = None

    # get the exception data
    name = ex.__class__.__name__

    # get level
    level = logging.getLevelName(level)

    # log the exception
    if EXCEPTION_TRACE:
        logger.log(level, "exception : " + name, exc_info=ex)
    else:
        logger.log(level, "exception : " + name + "\n" + str(ex))


def get_timer():
    """
    Get a timestamp with the current time.

    Returns
    -------
    timestamp : timestamp
        Timestamp with the current time.
    """

    timestamp = datetime.datetime.today()

    return timestamp


def get_duration(timestamp):
    """
    Get the elapsed time with respect to a timestamp.

    Parameters
    ----------
    timestamp : timestamp
        Timestamp with the reference time.

    Returns
    -------
    seconds : float
        Float with the elapsed time in seconds.
    duration : string
        String with the formatted elapsed time.
    date : string
        String with the formatted initial timestamp.
    """

    # get timing
    duration = datetime.datetime.today()-timestamp

    # parse timing
    seconds = duration.total_seconds()
    duration = _get_format_duration(duration)
    date = _get_format_timestamp(timestamp)

    return seconds, duration, date


def set_global(timestamp, level):
    """
    Set the global variables.
        - timestamp (for the elapsed time)
        - indentation level (for log messages)

    Parameters
    ----------
    timestamp : timestamp
        Timestamp (for the elapsed time).
    level : integer
        Indentation level for the log messages.
    """

    global GLOBAL_TIMESTAMP
    global GLOBAL_LEVEL
    GLOBAL_TIMESTAMP = timestamp
    GLOBAL_LEVEL = level


def get_global():
    """
    Get the global variables.
        - timestamp (for the elapsed time)
        - indentation level (for log messages)

    Returns
    -------
    timestamp : timestamp
        Timestamp (for the elapsed time).
    level : integer
        Indentation level for the log messages.
    """

    return GLOBAL_TIMESTAMP, GLOBAL_LEVEL


def get_logger(name):
    """
    Get a logger with a specified name.

    Parameters
    ----------
    name : string
        Name of the logger to be returned.
        If the logger does not exist, the logger is created.
        If the logger does exist, the logger is returned.

    Returns
    -------
    logger : logger
        Logger object instance.
    """

    # get the logger
    logger = logging.getLogger(name)

    # create the logger (if not already done)
    if len(logger.handlers) == 0:
        # get the formatter
        fmt = _DeltaTimeFormatter()

        # get the handle
        handler = logging.StreamHandler()
        handler.setFormatter(fmt)

        # prevent duplicated log messages
        logger.propagate = False

        # set the level
        logger.setLevel(LEVEL)

        # get the logger
        logger.addHandler(handler)

    return logger


# load the config file
try:
    # load the default file
    with importlib.resources.path("pypeec.data", "logger.yaml") as file:
        with open(file, 'r') as fid:
            data = yaml.safe_load(fid)

    # get file custom file
    file = os.getenv("PYTHONLOGGER")
    if file is not None:
        with open(file, 'r') as fid:
            data = yaml.safe_load(fid)

    # check file integrity
    _check_config(data)

    # set global variables
    FORMAT = data["FORMAT"]
    LEVEL = data["LEVEL"]
    INDENTATION = data["INDENTATION"]
    EXCEPTION_TRACE = data["EXCEPTION_TRACE"]
    USE_COLOR = data["USE_COLOR"]
    DEF_COLOR = data["DEF_COLOR"]
except Exception as ex:
    print("==========================", file=sys.stderr)
    print("INVALID CONFIGURATION FILE", file=sys.stderr)
    print("==========================", file=sys.stderr)
    print(str(ex), file=sys.stderr)
    print("==========================", file=sys.stderr)
    sys.exit(1)



"""
Module containing the command line script.
    - Used for calling the mesher, solver, viewer, and plotter.
    - Parse the command line arguments.
    - Call the corresponding entry point.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import sys
import argparse
import importlib.resources
from pypeec import main


def _get_parser():
    """
    Create a command line parser with a description.
    """

    # get the parser
    parser = argparse.ArgumentParser(
        prog="pypeec",
        description="PyPEEC - 3D PEEC Solver",
        epilog="Thomas Guillod - Dartmouth College",
        allow_abbrev=False,
    )

    # get version
    try:
        with importlib.resources.open_text("pypeec.data", "version.txt") as file_version:
            version = file_version.read()
    except FileNotFoundError:
        version = 'x.x.x'

    # display the version
    parser.add_argument(
        "-v", "--version",
        action="version",
        version="PyPEEC %s" % version,
    )

    # add subparsers
    subparsers = parser.add_subparsers(
        required=True,
        dest="command",
        title="command",
    )

    return parser, subparsers


def _get_arg_mesher(subparsers):
    """
    Add the mesher arguments.
    """

    # add the subparser
    parser = subparsers.add_parser(
        "mesher",
        aliases=["me"],
        help="mesh a geometry into a 3D voxel structure",
    )

    # add the arguments
    parser.add_argument(
        "-ge", "--geometry",
        help="geometry file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_geometry",
    )
    parser.add_argument(
        "-vo", "--voxel",
        help="voxel file (output / pickle)",
        required=True,
        metavar="file",
        dest="file_voxel",
    )
    parser.add_argument(
        "-t", "--truncated",
        help="truncate the results (default: full results)",
        action="store_true",
        dest="is_truncated",
    )


def _get_arg_viewer(subparsers):
    """
    Add the viewer arguments.
    """

    # add the subparser
    parser = subparsers.add_parser(
        "viewer",
        aliases=["vi"],
        help="visualization of a 3D voxel structure",
    )

    # add the arguments
    parser.add_argument(
        "-vo", "--voxel",
        help="voxel file (input / pickle)",
        required=True,
        metavar="file",
        dest="file_voxel",
    )
    parser.add_argument(
        "-vi", "--viewer",
        help="viewer file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_viewer",
    )
    parser.add_argument(
        "-tp", "--tag_plot",
        help="list of plots to be shown (default: all the plots)",
        nargs='+',
        default=None,
        metavar="tag_plot",
        dest="tag_plot",
    )
    parser.add_argument(
        "-pm", "--plot_mode",
        help="selection of the plot mode (default: window)",
        choices=["qt", "nb", "save", "none"],
        default="qt",
        dest="plot_mode",
    )
    parser.add_argument(
        "-f", "--folder",
        help="folder for saving the screenshots (default: cwd)",
        default=".",
        metavar="folder",
        dest="folder",
    )


def _get_arg_solver(subparsers):
    """
    Add the solver arguments.
    """

    # add the subparser
    parser = subparsers.add_parser(
        "solver",
        aliases=["so"],
        help="solve a problem with the PEEC method",
    )

    # add the arguments
    parser.add_argument(
        "-vo", "--voxel",
        help="voxel file (input / pickle)",
        required=True,
        metavar="file",
        dest="file_voxel",
    )
    parser.add_argument(
        "-pr", "--problem",
        help="problem file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_problem",
    )
    parser.add_argument(
        "-to", "--tolerance",
        help="tolerance file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_tolerance",
    )
    parser.add_argument(
        "-so", "--solution",
        help="solution file (output / pickle)",
        required=True,
        metavar="file",
        dest="file_solution",
    )
    parser.add_argument(
        "-t", "--truncated",
        help="truncate the results (default: full results)",
        action="store_true",
        dest="is_truncated",
    )


def _get_arg_plotter(subparsers):
    """
    Add the plotter arguments.
    """

    # add the subparser
    parser = subparsers.add_parser(
        "plotter",
        aliases=["pl"],
        help="plot the solution of a PEEC problem",
    )

    # add the arguments
    parser.add_argument(
        "-so", "--solution",
        help="solution file (input / pickle)",
        required=True,
        metavar="file",
        dest="file_solution",
    )
    parser.add_argument(
        "-pl", "--plotter",
        help="plotter file (input / JSON or YAML)",
        required=True,
        metavar="file",
        dest="file_plotter",
    )
    parser.add_argument(
        "-ts", "--tag_sweep",
        help="list of sweeps to be shown (default: show the sweeps)",
        nargs='+',
        default=None,
        metavar="tag_sweep",
        dest="tag_sweep",
    )
    parser.add_argument(
        "-tp", "--tag_plot",
        help="list of plots to be shown (default: all the plots)",
        nargs='+',
        default=None,
        metavar="tag_plot",
        dest="tag_plot",
    )
    parser.add_argument(
        "-pm", "--plot_mode",
        help="selection of the plot mode (default: window)",
        choices=["qt", "nb", "save", "none"],
        default="qt",
        dest="plot_mode",
    )
    parser.add_argument(
        "-f", "--folder",
        help="folder for saving the screenshots (default: cwd)",
        default=".",
        metavar="folder",
        dest="folder",
    )


def _get_arg_extract(subparsers):
    """
    Add the data extraction arguments (examples and documentation).
    """

    # add the examples data parser
    parser = subparsers.add_parser(
        "examples",
        help="extract the examples",
    )
    parser.add_argument(
        help="path where the examples should be extracted",
        type=str,
        metavar="path",
        dest="path_extract",
    )

    # add the documentation data parser
    parser = subparsers.add_parser(
        "documentation",
        help="extract the documentation",
    )
    parser.add_argument(
        help="path where the documentation should be extracted",
        type=str,
        metavar="path",
        dest="path_extract",
    )

    # add the logger config parser
    parser = subparsers.add_parser(
        "logger",
        help="extract the default logger config file",
    )
    parser.add_argument(
        help="path where the logger config file should be extracted",
        type=str,
        metavar="path",
        dest="path_extract",
    )


def run_arguments(argv):
    """
    User script for running PyPEEC with given arguments.
        - The script offers a CLI for the mesher, solver, viewer, and plotter.
        - The script can also be used to extract data (examples, documentation, and logger config).

    Parameters
    ----------
    argv : list
        List with the command line arguments.

    Returns
    -------
    status : int
        The status exit code of the script.

    """

    # get the parser
    (parser, subparsers) = _get_parser()

    # add the sub-command arguments
    _get_arg_mesher(subparsers)
    _get_arg_viewer(subparsers)
    _get_arg_solver(subparsers)
    _get_arg_plotter(subparsers)
    _get_arg_extract(subparsers)

    # parse the config and get arguments
    try:
        args = parser.parse_args(argv)
    except SystemExit as status:
        return status.code

    # run the code
    try:
        if args.command in ["mesher", "me"]:
            main.run_mesher_file(
                args.file_geometry,
                args.file_voxel,
                is_truncated=args.is_truncated,
            )
        elif args.command in ["viewer", "vi"]:
            main.run_viewer_file(
                args.file_voxel,
                args.file_viewer,
                tag_plot=args.tag_plot,
                plot_mode=args.plot_mode,
                folder=args.folder,
            )
        elif args.command in ["solver", "so"]:
            main.run_solver_file(
                args.file_voxel,
                args.file_problem,
                args.file_tolerance,
                args.file_solution,
                is_truncated=args.is_truncated,
            )
        elif args.command in ["plotter", "pl"]:
            main.run_plotter_file(
                args.file_solution,
                args.file_plotter,
                tag_sweep=args.tag_sweep,
                tag_plot=args.tag_plot,
                plot_mode=args.plot_mode,
                folder=args.folder,
            )
        elif args.command == "examples":
            main.run_extract("examples.zip", True, args.path_extract)
        elif args.command == "documentation":
            main.run_extract("documentation.zip", True, args.path_extract)
        elif args.command == "logger":
            main.run_extract("logger.yaml", False, args.path_extract)
        else:
            raise ValueError("invalid command")
    except Exception:
        return 1
    else:
        return 0


def run_script():
    """
    User script for running PyPEEC with the command line arguments.
        - The script offers a CLI for the mesher, solver, viewer, and plotter.
        - The script can also be used to extract data (examples, documentation, and logger config).

    The script is installed with the package.
    The name of the command line script is "pypeec".
    """

    # get arguments
    argv = sys.argv[1:]

    # run the script
    status = run_arguments(argv)

    # exit code
    sys.exit(status)


if __name__ == "__main__":
    run_script()



"""
Module for determining and exploiting the symmetry between terminals.

Determine which simulations are required to extract the full impedance matrix.
This means that a reduced set of simulations are performed.
Finally, the full solution is expanded from the reduced set.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np


def _get_excitation_compact(data, symmetry):
    """
    Reduce a matrix with respect to a particular symmetry.
    Only a single excitation per symmetry is kept.
    """

    # extract
    perm = symmetry["perm"]
    ref = symmetry["ref"]

    # cast and check
    perm = np.array(perm, dtype=int)
    if ref >= len(perm):
        raise ValueError("invalid index")

    # combine excitations with the symmetry
    for perm_tmp in perm:
        data[perm_tmp] = data[perm[ref]]

    return data


def _get_solution_expand(n_winding, data, symmetry):
    """
    Expand a matrix with respect to a particular symmetry.
    Different solutions are generated with permutations.
    The permutations are given by the symmetry.
    """

    # extract
    perm = symmetry["perm"]
    ref = symmetry["ref"]

    # cast and check
    perm = np.array(perm, dtype=int)
    if ref >= len(perm):
        raise ValueError("invalid index")

    # initialize the expanded matrix
    data_all = np.zeros((n_winding, 0), dtype=data.dtype)

    # index for the permutation
    idx = np.arange(len(perm))

    # combine excitations
    for i in range(len(perm)):
        # get the permutation to be performed
        idx_add = np.roll(idx, i)
        perm_add = perm[idx_add, :]

        # permute the matrix
        data_add = np.copy(data)
        data_add[perm.flatten(), :] = data[perm_add.flatten(), :]

        # add the permuted matrix the expanded matrix
        data_all = np.hstack((data_all, data_add))

    return data_all


def get_excitation_all(n_winding, symmetry):
    """
    Determine which simulations are required to extract the full impedance matrix.
    The different symmetries are used to reduce the number of simulations.
    """

    # full excitations
    excitation = np.arange(n_winding)

    # combine excitations with the symmetries
    for symmetry_tmp in symmetry:
        excitation = _get_excitation_compact(excitation, symmetry_tmp)

    # remove redundant excitations
    excitation = np.unique(excitation)

    # get the number of solutions for the reduced problem
    n_solution = len(excitation)

    return n_solution, excitation


def get_solution_all(terminal, symmetry):
    """
    Generate a full solution from a reduced solution.
    The different symmetries are used to expand the number of simulations.
    """

    # extract
    I_mat = terminal["I_mat"]
    V_mat = terminal["V_mat"]
    n_winding = terminal["n_winding"]
    n_solution = terminal["n_solution"]
    freq = terminal["freq"]

    # check size
    assert I_mat.shape == (n_winding, n_solution), "invalid solution: current matrix shape"
    assert V_mat.shape == (n_winding, n_solution), "invalid solution: voltage matrix shape"

    # expand the solution with the symmetries
    for sym_tmp in symmetry:
        I_mat = _get_solution_expand(n_winding, I_mat, sym_tmp)
        V_mat = _get_solution_expand(n_winding, V_mat, sym_tmp)

    # remove redundant solutions
    IV_mat = np.vstack((I_mat, V_mat))
    (n_terminal, n_solution) = IV_mat.shape
    assert n_terminal == (2*n_winding), "invalid solution: terminal matrix shape"

    # update terminal
    terminal = {
        "n_winding": n_winding,
        "n_solution": n_solution,
        "freq": freq,
        "I_mat": I_mat,
        "V_mat": V_mat,
    }

    return terminal



"""
Different functions for plotting voxel structures with PyVista.

For the viewer and the plotter, the following object are shown:
    - the complete voxel structure (as wireframe)
    - the structure containing non-empty voxels (as wireframe)
    - the defined point cloud (as points)

For the viewer, the following plots are available:
    - the domains are shown for the non-empty voxels
    - the connected components for the non-empty voxels

For the plotter, the following plots are available:
    - material description for the non-empty voxels
    - scalar plots for the non-empty voxels
    - arrow plots for the non-empty voxels
    - scalar plots for the point cloud
    - arrow plots for the point cloud
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import numpy.linalg as lna


def _get_plot_view_theme(pl, grid, voxel, point, plot_view, plot_theme):
    """
    Plot the geometry as wireframe (complete grid and non-empty voxels).
    Plot the point cloud used for the field evaluation.
    Add the axis descriptor to orientate the geometry.
    Set the plot theme.
    """

    # set background
    pl.set_background(plot_theme["background_color"])

    # plot the complete grid
    if plot_view["grid_plot"] and (grid.n_cells > 0):
        pl.add_mesh(
            grid,
            style="wireframe",
            color=plot_view["grid_color"],
            opacity=plot_view["grid_opacity"],
            line_width=plot_view["grid_thickness"]
        )

    # plot the non-empty voxels
    if plot_view["geom_plot"] and (voxel.n_cells > 0):
        pl.add_mesh(
            voxel,
            style="wireframe",
            color=plot_view["geom_color"],
            opacity=plot_view["geom_opacity"],
            line_width=plot_view["geom_thickness"]
        )

    # plot the cloud points
    if plot_view["point_plot"] and (point.n_cells > 0):
        pl.add_mesh(
            point,
            color=plot_view["point_color"],
            point_size=plot_view["point_size"],
            opacity=plot_view["point_opacity"],
            render_points_as_spheres=True,
        )

    # set the camera position
    if plot_view["camera_roll"] is not None:
        pl.camera.roll = plot_view["camera_roll"]
    if plot_view["camera_azimuth"] is not None:
        pl.camera.azimuth = plot_view["camera_azimuth"]
    if plot_view["camera_elevation"] is not None:
        pl.camera.elevation = plot_view["camera_elevation"]

    # add axes
    if plot_theme["axis_add"]:
        pl.add_axes(
            line_width=plot_theme["axis_size"],
            color=plot_theme["text_color"],
            interactive=False,
        )


def _get_plot_title(pl, title, plot_theme):
    """
    Add a title located at the corner of the plot.
    """

    # add titles
    pl.add_text(
        title,
        font_size=plot_theme["title_font"],
        color=plot_theme["text_color"],
    )


def _get_clip_mesh(pl, obj, arg, plot_clip):
    """
    Add an object (either full view or clipped).
    """

    # extract
    clip_plot = plot_clip["clip_plot"]
    clip_invert = plot_clip["clip_invert"]
    clip_axis = plot_clip["clip_axis"]
    clip_value = plot_clip["clip_value"]

    # add the plot
    if clip_plot:
        # add the clipping arguments
        arg["invert"] = clip_invert
        arg["normal"] = clip_axis
        arg["value"] = clip_value
        arg["normal_rotation"] = False

        # add the clipped plot
        pl.add_mesh_clip_plane(
            obj,
            **arg,
        )
    else:
        # add the full plot
        pl.add_mesh(
            obj,
            **arg,
        )


def _get_phase_vector(obj, var, phase):
    """
    Get a vector for a specified phase shift.
    The vector is constructed from the phasor.
    """

    # get var
    data_re = obj[var + "_re"]
    data_im = obj[var + "_im"]

    # construct the vector
    time = data_re+1j*data_im

    # phase shift the vector
    time *= np.exp(1j*phase)

    # get the vector in the time domain
    time = np.real(time)

    # get the norm
    norm = lna.norm(time, axis=1)

    # assign the data
    obj[var + "_time"] = time
    obj[var + "_norm"] = norm

    return obj


def _get_filter_vector(obj, var, arrow_threshold):
    """
    Filter the voxel structure with a vector field.
    This function is used to remove arrows with extremely low lengths.
    """

    # if the voxel structure is empty, nothing to do
    if obj.n_cells == 0:
        return obj

    # get var
    data = obj[var]

    # threshold for arrow removal
    if np.any(np.isfinite(data)):
        thr = np.nanmax(data)*arrow_threshold
    else:
        thr = np.nan

    # indices to be kept
    idx = data > thr

    # filter out the arrows that are too small
    obj = obj.extract_cells(idx)

    return obj


def _get_scale_scalar(obj, var, scale):
    """
    Scale a scalar variable between a lower and upper bound.
    """

    # if the voxel structure is empty, nothing to do
    if obj.n_cells == 0:
        return obj

    # get var
    data = obj[var]

    # add scaling
    data = scale*data

    # assign data
    obj[var] = data

    return obj


def _get_filter_scalar(obj, var, filter_lim):
    """
    Filter the voxel structure with provided limits with respect to a scalar variable.
    """

    # if the voxel structure is empty, nothing to do
    if obj.n_cells == 0:
        return obj

    # handle None
    if filter_lim is None:
        f_min = -float("inf")
        f_max = +float("inf")
    else:
        (f_min, f_max) = filter_lim

    # get var
    data = obj[var]

    # indices to be kept
    idx = np.logical_and(data >= f_min, data <= f_max)

    # filter data
    obj = obj.extract_cells(idx)

    return obj


def _get_clamp_scalar(obj, var, clamp_lim):
    """
    Clamp a scalar variable between a lower and upper bound.
    """

    # if the voxel structure is empty, nothing to do
    if obj.n_cells == 0:
        return obj

    # handle None
    if clamp_lim is None:
        c_min = -float("inf")
        c_max = +float("inf")
    else:
        (c_min, c_max) = clamp_lim

    # get var
    data = obj[var]

    # clamp range
    data = np.maximum(data, c_min)
    data = np.minimum(data, c_max)

    # assign data
    obj[var] = data

    return obj


def _plot_scalar(pl, obj, data_plot, plot_clip, plot_theme):
    """
    Plot a scalar variable.
    The plot is either made on:
        - the unstructured grid describing the non-empty voxels
        - the polydata (point cloud) used to evaluate the field
    """

    # extract
    var = data_plot["var"]
    scale = data_plot["scale"]
    log = data_plot["log"]
    filter_lim = data_plot["filter_lim"]
    clamp_lim = data_plot["clamp_lim"]
    color_lim = data_plot["color_lim"]
    point_size = data_plot["point_size"]
    legend = data_plot["legend"]
    title = data_plot["title"]

    # extract
    colorbar_plot = plot_theme["colorbar_plot"]
    colorbar_size = plot_theme["colorbar_size"]
    colorbar_font = plot_theme["colorbar_font"]
    text_color = plot_theme["text_color"]

    # set title
    _get_plot_title(pl, title, plot_theme)

    # color bar options
    scalar_bar_args = dict(
        title=legend,
        n_labels=colorbar_size,
        label_font_size=colorbar_font,
        title_font_size=colorbar_font,
        color=text_color,
    )

    # scale and clamp the variable
    obj_tmp = obj.copy(deep=True)
    obj_tmp = _get_scale_scalar(obj_tmp, var, scale)
    obj_tmp = _get_filter_scalar(obj_tmp, var, filter_lim)
    obj_tmp = _get_clamp_scalar(obj_tmp, var, clamp_lim)

    # add the resulting plot to the plotter
    arg = dict(
        scalars=var,
        log_scale=log,
        clim=color_lim,
        point_size=point_size,
        show_scalar_bar=colorbar_plot,
        scalar_bar_args=scalar_bar_args,
        render_points_as_spheres=True,
    )
    if obj_tmp.n_cells > 0:
        _get_clip_mesh(pl, obj_tmp, arg, plot_clip)


def _plot_arrow(pl, grid, obj, data_plot, plot_clip, plot_theme):
    """
    Plot a vector variable with an arrow plot (quiver plot).
    The plot is either made on:
        - the unstructured grid describing the non-empty voxels
        - the polydata (point cloud) used to evaluate the field

    A scalar variable is used to determine the color of the arrows.
    The length of the arrows is constant (and scaled with respect to the voxel size).
    """

    # extract
    var = data_plot["var"]
    phase = data_plot["phase"]
    scale = data_plot["scale"]
    log = data_plot["log"]
    filter_lim = data_plot["filter_lim"]
    clamp_lim = data_plot["clamp_lim"]
    color_lim = data_plot["color_lim"]
    arrow_scale = data_plot["arrow_scale"]
    arrow_threshold = data_plot["arrow_threshold"]
    legend = data_plot["legend"]
    title = data_plot["title"]

    # extract
    colorbar_plot = plot_theme["colorbar_plot"]
    colorbar_size = plot_theme["colorbar_size"]
    colorbar_font = plot_theme["colorbar_font"]
    text_color = plot_theme["text_color"]

    # set title
    _get_plot_title(pl, title, plot_theme)

    # color bar options
    scalar_bar_args = dict(
        title=legend,
        n_labels=colorbar_size,
        label_font_size=colorbar_font,
        title_font_size=colorbar_font,
        color=text_color,
    )

    # get variable name
    var_time = var + "_time"
    var_norm = var + "_norm"

    # scale and clamp the variable
    obj_tmp = obj.copy(deep=True)
    obj_tmp = _get_phase_vector(obj_tmp, var, phase)
    obj_tmp = _get_filter_vector(obj_tmp, var_norm, arrow_threshold)
    obj_tmp = _get_scale_scalar(obj_tmp, var_norm, scale)
    obj_tmp = _get_filter_scalar(obj_tmp, var_norm, filter_lim)
    obj_tmp = _get_clamp_scalar(obj_tmp, var_norm, clamp_lim)

    # get arrow size
    d_char = min(grid.spacing)
    factor = d_char*arrow_scale

    # add the resulting plot to the plotter
    arg = dict(
        scalars=var_norm,
        log_scale=log,
        clim=color_lim,
        show_scalar_bar=colorbar_plot,
        scalar_bar_args=scalar_bar_args,
    )
    if obj_tmp.n_cells > 0:
        glyph_tmp = obj_tmp.glyph(orient=var_time, scale=False, factor=factor)
        _get_clip_mesh(pl, glyph_tmp, arg, plot_clip)


def _plot_material(pl, voxel, data_plot, plot_clip, plot_theme):
    """
    Plot the material and source description.
    """

    # extract
    color_electric = data_plot["color_electric"]
    color_magnetic = data_plot["color_magnetic"]
    color_current_source = data_plot["color_current_source"]
    color_voltage_source = data_plot["color_voltage_source"]
    title = data_plot["title"]

    # set title
    _get_plot_title(pl, title, plot_theme)

    # get a colormap with three discrete color
    cmap = [
        color_electric,
        color_magnetic,
        color_current_source,
        color_voltage_source,
    ]

    # make a copy (for avoid cross coupling)
    voxel_tmp = voxel.copy(deep=True)

    # add the resulting plot to the plotter
    arg = dict(
        clim=[1, 4],
        show_scalar_bar=False,
        scalars="material",
        cmap=cmap,
    )
    if voxel_tmp.n_cells > 0:
        _get_clip_mesh(pl, voxel_tmp, arg, plot_clip)


def _plot_geometry(pl, voxel, data_plot, plot_clip, plot_theme, tag):
    """
    Plot an integer variable on the voxel structure (material or connection).
    """

    # extract
    colormap = data_plot["colormap"]
    opacity = data_plot["opacity"]
    title = data_plot["title"]

    # set title
    _get_plot_title(pl, title, plot_theme)

    # make a copy (for avoid cross coupling)
    voxel_tmp = voxel.copy(deep=True)

    # add the resulting plot to the plotter
    arg = dict(
        show_scalar_bar=False,
        scalars=tag,
        cmap=colormap,
        opacity=opacity,
    )
    if voxel_tmp.n_cells > 0:
        _get_clip_mesh(pl, voxel_tmp, arg, plot_clip)


def _plot_voxelization(pl, voxel, reference, data_plot, plot_clip, plot_theme):
    """
    Plot the reference and voxelized structures in order to assess the voxelization error.
    """

    # extract the data
    color_voxel = data_plot["color_voxel"]
    color_reference = data_plot["color_reference"]
    opacity_voxel = data_plot["opacity_voxel"]
    opacity_reference = data_plot["opacity_reference"]
    title = data_plot["title"]

    # set title
    _get_plot_title(pl, title, plot_theme)

    # make a copy (for avoid cross coupling)
    voxel_tmp = voxel.copy(deep=True)
    reference_tmp = reference.copy(deep=True)

    # add the resulting plot to the plotter
    arg = dict(
        show_scalar_bar=False,
        color=color_voxel,
        opacity=opacity_voxel,
    )
    if voxel_tmp.n_cells > 0:
        _get_clip_mesh(pl, voxel_tmp, arg, plot_clip)

    # add the resulting plot to the plotter
    arg = dict(
        show_scalar_bar=False,
        color=color_reference,
        opacity=opacity_reference,
    )
    if voxel_tmp.n_cells > 0:
        _get_clip_mesh(pl, reference_tmp, arg, plot_clip)


def get_plot_viewer(pl, grid, voxel, point, reference, layout, data_plot, data_options):
    """
    Plot the voxel structure (for the viewer).
    The following plot types are available:
        - the domains are shown for the non-empty voxels
        - the connected components for the non-empty voxels
        - the meshing tolerance between the reference and voxelized structures
    """

    # get options
    plot_clip = data_options["plot_clip"]
    plot_view = data_options["plot_view"]
    plot_theme = data_options["plot_theme"]

    # get the main plot
    if layout == "domain":
        _plot_geometry(pl, voxel, data_plot, plot_clip, plot_theme, "domain")
    elif layout == "connection":
        _plot_geometry(pl, voxel, data_plot, plot_clip, plot_theme, "connection")
    elif layout == "voxelization":
        _plot_voxelization(pl, voxel, reference, data_plot, plot_clip, plot_theme)
    else:
        raise ValueError("invalid plot type and plot feature")

    # add the wireframe and axis
    _get_plot_view_theme(pl, grid, voxel, point, plot_view, plot_theme)


def get_plot_plotter(pl, grid, voxel, point, layout, data_plot, data_options):
    """
    Plot the solution (for the plotter).
    The following plot types are available:
        - plot the material and source description on the voxel structure
        - plot a scalar variable on the voxel structure
        - plot a scalar variable on the point cloud
        - plot a vector variable on the voxel structure
        - plot a vector variable on the point cloud
    """

    # extract the data
    plot_clip = data_options["plot_clip"]
    plot_view = data_options["plot_view"]
    plot_theme = data_options["plot_theme"]

    # get the main plot
    if layout == "material":
        _plot_material(pl, voxel, data_plot, plot_clip, plot_theme)
    elif layout == "scalar_voxel":
        _plot_scalar(pl, voxel, data_plot, plot_clip, plot_theme)
    elif layout == "scalar_point":
        _plot_scalar(pl, point, data_plot, plot_clip, plot_theme)
    elif layout == "arrow_voxel":
        _plot_arrow(pl, grid, voxel, data_plot, plot_clip, plot_theme)
    elif layout == "arrow_point":
        _plot_arrow(pl, grid, point, data_plot, plot_clip, plot_theme)
    else:
        raise ValueError("invalid plot type and plot feature")

    # add the wireframe and axis
    _get_plot_view_theme(pl, grid, voxel, point, plot_view, plot_theme)



"""
Module for checking the viewer and plotter data.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_check import datachecker


def _check_data_window(data_window):
    """
    Check the plot window options (for the viewer and plotter).
    Check the window size and menu configuration.
    """

    # check type
    key_list = ["show_menu", "image_size", "window_size", "notebook_size"]
    datachecker.check_dict("data_window", data_window, key_list=key_list)

    # extract the data
    show_menu = data_window["show_menu"]
    image_size = data_window["image_size"]
    window_size = data_window["window_size"]
    notebook_size = data_window["notebook_size"]

    # check data
    datachecker.check_boolean("show_menu", show_menu)
    datachecker.check_integer_array("image_size", image_size, size=2, is_positive=True, can_be_zero=False, can_be_none=True)
    datachecker.check_integer_array("window_size", window_size, size=2, is_positive=True, can_be_zero=False, can_be_none=True)
    datachecker.check_integer_array("notebook_size", notebook_size, size=2, is_positive=True, can_be_zero=False, can_be_none=True)


def _check_data_options_matplotlib(data_options):
    """
    Check the plot options (for Matplotlib).
    """

    # check type
    key_list = ["style", "legend", "font"]
    datachecker.check_dict("data_options", data_options, key_list=key_list)

    # extract the data
    style = data_options["style"]
    legend = data_options["legend"]
    font = data_options["font"]

    # check data
    datachecker.check_string("style", style, can_be_empty=False)
    datachecker.check_string("legend", legend, can_be_empty=False)
    datachecker.check_integer("font", font, is_positive=True, can_be_zero=False)


def _check_data_options_pyvista(data_options):
    """
    Check the plot options (for PyVista).
    """

    # check type
    key_list = ["plot_clip", "plot_view", "plot_theme"]
    datachecker.check_dict("data_options", data_options, key_list=key_list)

    # extract the data
    plot_clip = data_options["plot_clip"]
    plot_view = data_options["plot_view"]
    plot_theme = data_options["plot_theme"]

    _check_plot_clip(plot_clip)
    _check_plot_view(plot_view)
    _check_plot_theme(plot_theme)


def _check_plot_view(plot_view):
    """
    Check the validity of the plot view (for PyVista).
    """

    # check type
    key_list = [
        "grid_plot", "grid_thickness", "grid_color", "grid_opacity",
        "geom_plot", "geom_thickness", "geom_color", "geom_opacity",
        "point_plot", "point_size", "point_color", "point_opacity",
        "camera_roll", "camera_azimuth", "camera_elevation",
    ]
    datachecker.check_dict("data_window", plot_view, key_list=key_list)

    # check grid data
    datachecker.check_boolean("grid_plot", plot_view["grid_plot"])
    datachecker.check_string("grid_color", plot_view["grid_color"], can_be_empty=False)
    datachecker.check_float("grid_thickness", plot_view["grid_thickness"], is_positive=True, can_be_zero=True)
    datachecker.check_float("grid_opacity", plot_view["grid_opacity"], is_positive=True, can_be_zero=True)

    # check geom data
    datachecker.check_boolean("geom_plot", plot_view["geom_plot"])
    datachecker.check_string("geom_color", plot_view["geom_color"], can_be_empty=False)
    datachecker.check_float("geom_thickness", plot_view["geom_thickness"], is_positive=True, can_be_zero=True)
    datachecker.check_float("geom_opacity", plot_view["geom_opacity"], is_positive=True, can_be_zero=True)

    # check cloud data
    datachecker.check_boolean("point_plot", plot_view["point_plot"])
    datachecker.check_string("point_color", plot_view["point_color"], can_be_empty=False)
    datachecker.check_float("point_size", plot_view["point_size"], is_positive=True, can_be_zero=True)
    datachecker.check_float("point_opacity", plot_view["point_opacity"], is_positive=True, can_be_zero=True)

    # check camera data
    datachecker.check_float("camera_roll", plot_view["camera_roll"], can_be_none=True)
    datachecker.check_float("camera_azimuth", plot_view["camera_azimuth"], can_be_none=True)
    datachecker.check_float("camera_elevation", plot_view["camera_elevation"], can_be_none=True)


def _check_plot_theme(plot_theme):
    """
    Check the validity of the plot theme (for PyVista).
    """

    # check type
    key_list = [
        "text_color", "title_font",
        "colorbar_font", "colorbar_size", "colorbar_plot",
        "axis_add", "axis_size",
        "background_color",
    ]
    datachecker.check_dict("data_window", plot_theme, key_list=key_list)

    # check data
    datachecker.check_boolean("axis_add", plot_theme["axis_add"])
    datachecker.check_boolean("colorbar_plot", plot_theme["colorbar_plot"])
    datachecker.check_string("text_color", plot_theme["text_color"], can_be_empty=False)
    datachecker.check_string("background_color", plot_theme["background_color"], can_be_empty=False)
    datachecker.check_integer("title_font", plot_theme["title_font"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("colorbar_font", plot_theme["colorbar_font"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("colorbar_size", plot_theme["colorbar_size"], is_positive=True, can_be_zero=False)
    datachecker.check_integer("axis_size", plot_theme["axis_size"], is_positive=True, can_be_zero=False)


def _check_plot_clip(plot_clip):
    """
    Check the validity of the plot cut plane (for PyVista).
    """

    # check type
    key_list = ["clip_plot", "clip_invert", "clip_axis", "clip_value"]
    datachecker.check_dict("plot_clip", plot_clip, key_list=key_list)

    # check data
    datachecker.check_boolean("clip_plot", plot_clip["clip_plot"])
    datachecker.check_boolean("clip_invert", plot_clip["clip_invert"])
    datachecker.check_choice("clip_axis", plot_clip["clip_axis"], ["x", "y", "z"])
    datachecker.check_float("clip_value", plot_clip["clip_value"])


def _check_data_plot_plotter(layout, data_plot):
    """
    Check the validity of the data options (for the plotter).
    The data options are controlling the plot content.
    """

    # list of allowed plot types
    layout_list = [
        "convergence",
        "residuum",
        "material",
        "scalar_voxel",
        "scalar_point",
        "arrow_voxel",
        "arrow_point",
    ]

    # list of allowed variable names
    var_voxel_list = [
        "V_c_re", "V_c_im", "V_c_abs",
        "V_m_re", "V_m_im", "V_m_abs",
        "S_c_re", "S_c_im", "S_c_abs",
        "Q_m_re", "Q_m_im", "Q_m_abs",
        "J_c_norm", "B_m_norm",
        "P_c_abs", "P_m_abs",
    ]
    var_point_list = ["H_norm"]
    vec_point_list = ["H_vec"]
    vec_voxel_list = ["J_c_vec", "B_m_vec"]

    # check plot type
    datachecker.check_choice("layout", layout, layout_list)

    # check the convergence options
    if layout == "convergence":
        # check type
        key_list = ["color_active", "color_reactive", "marker"]
        datachecker.check_dict("data_plot", data_plot, key_list=key_list)

        # check data
        datachecker.check_string("color_active", data_plot["color_active"], can_be_empty=False)
        datachecker.check_string("color_reactive", data_plot["color_reactive"], can_be_empty=False)
        datachecker.check_float("marker", data_plot["marker"], is_positive=True, can_be_zero=False)
        datachecker.check_float("width", data_plot["width"], is_positive=True, can_be_zero=False)

    # check the residuum options
    if layout == "residuum":
        # check type
        key_list = ["n_bins", "tol_bins", "edge_color", "bar_color"]
        datachecker.check_dict("data_plot", data_plot, key_list=key_list)

        # check data
        datachecker.check_integer("n_bins", data_plot["n_bins"], is_positive=True, can_be_zero=False)
        datachecker.check_float("tol_bins", data_plot["tol_bins"], is_positive=True, can_be_zero=False)
        datachecker.check_string("edge_color", data_plot["edge_color"], can_be_empty=False)
        datachecker.check_string("bar_color", data_plot["bar_color"], can_be_empty=False)

    # check the material options
    if layout == "material":
        # check type
        key_list = ["title", "color_electric", "color_magnetic", "color_current_source", "color_voltage_source"]
        datachecker.check_dict("data_plot", data_plot, key_list=key_list)

        # check data
        datachecker.check_string("title", data_plot["title"], can_be_empty=True, can_be_none=True)
        datachecker.check_string("color_electric", data_plot["color_electric"], can_be_empty=False)
        datachecker.check_string("color_magnetic", data_plot["color_magnetic"], can_be_empty=False)
        datachecker.check_string("color_current_source", data_plot["color_current_source"], can_be_empty=False)
        datachecker.check_string("color_voltage_source", data_plot["color_voltage_source"], can_be_empty=False)

    # check the options for scalar and arrow plots
    if layout in ["scalar_voxel", "scalar_point", "arrow_voxel", "arrow_point"]:
        # check type
        key_list = ["title", "scale", "log", "legend", "color_lim", "clamp_lim", "filter_lim"]
        datachecker.check_dict("data_plot", data_plot, key_list=key_list)

        # check data
        datachecker.check_boolean("log", data_plot["log"])
        datachecker.check_string("title", data_plot["title"], can_be_empty=True, can_be_none=True)
        datachecker.check_string("legend", data_plot["legend"], can_be_empty=True, can_be_none=True)
        datachecker.check_float("scale", data_plot["scale"], is_positive=True, can_be_zero=True)
        datachecker.check_float_array("color_lim", data_plot["color_lim"], size=2, can_be_none=True)
        datachecker.check_float_array("clamp_lim", data_plot["clamp_lim"], size=2, can_be_none=True)
        datachecker.check_float_array("filter_lim", data_plot["filter_lim"], size=2, can_be_none=True)

    # check the scalar options
    if layout in ["scalar_voxel", "scalar_point"]:
        # check type
        key_list = ["var", "point_size"]
        datachecker.check_dict("data_plot", data_plot, key_list=key_list)

        # check data
        datachecker.check_string("var", data_plot["var"], can_be_empty=False)
        datachecker.check_float("point_size", data_plot["point_size"], is_positive=True, can_be_zero=True)

        # check compatibility
        if layout == "scalar_voxel":
            datachecker.check_choice("var", data_plot["var"], var_voxel_list)
        elif layout == "scalar_point":
            datachecker.check_choice("var", data_plot["var"], var_point_list)
        else:
            raise ValueError("plot_geom: the plot geometry option is incompatible with the plot type")

    # check the arrow options
    if layout in ["arrow_voxel", "arrow_point"]:
        # check type
        key_list = ["var", "phase", "arrow_scale", "arrow_threshold"]
        datachecker.check_dict("data_plot", data_plot, key_list=key_list)

        # check data
        datachecker.check_float("phase", data_plot["phase"])
        datachecker.check_string("var", data_plot["var"], can_be_empty=False)
        datachecker.check_float("arrow_scale", data_plot["arrow_scale"], is_positive=True, can_be_zero=True)
        datachecker.check_float("arrow_threshold", data_plot["arrow_threshold"], is_positive=True, can_be_zero=True)

        # check compatibility
        if layout == "arrow_voxel":
            datachecker.check_choice("var", data_plot["var"], vec_voxel_list)
        elif layout == "arrow_point":
            datachecker.check_choice("var", data_plot["var"], vec_point_list)
        else:
            raise ValueError("plot_geom: the plot geometry option is incompatible with the plot type")


def _check_data_plot_viewer(layout, data_plot):
    """
    Check the validity of the data options (for the viewer).
    The data options are controlling the plot content.
    """

    # check plot type
    datachecker.check_choice("layout", layout, ["domain", "connection", "voxelization"])

    # check the material options
    if layout == "voxelization":
        # check type
        key_list = ["color_voxel", "color_reference", "opacity_voxel", "opacity_reference"]
        datachecker.check_dict("data_plot", data_plot, key_list=key_list)

        # check data
        datachecker.check_string("title", data_plot["title"], can_be_empty=True, can_be_none=True)
        datachecker.check_string("color_voxel", data_plot["color_voxel"], can_be_empty=False)
        datachecker.check_string("color_reference", data_plot["color_reference"], can_be_empty=False)
        datachecker.check_float("opacity_voxel", data_plot["opacity_voxel"], is_positive=True, can_be_zero=False)
        datachecker.check_float("opacity_reference", data_plot["opacity_reference"], is_positive=True, can_be_zero=False)

    # check the scalar options
    if layout in ["domain", "connection"]:
        # check type
        key_list = ["colormap", "title", "opacity"]
        datachecker.check_dict("data_plot", data_plot, key_list=key_list)

        # check data
        datachecker.check_string("colormap", data_plot["colormap"], can_be_empty=False)
        datachecker.check_string("title", data_plot["title"], can_be_empty=True, can_be_none=True)
        datachecker.check_float("opacity", data_plot["opacity"], is_positive=True, can_be_zero=False)


def _check_data_item(data_item, handler):
    """
    Check the validity of the dict describing a single plot (for the viewer and plotter).
    """

    # check type
    key_list = ["title", "framework", "layout", "data_window", "data_plot", "data_options"]
    datachecker.check_dict("data_item", data_item, key_list=key_list)

    # extract field
    framework = data_item["framework"]
    title = data_item["title"]
    layout = data_item["layout"]
    data_window = data_item["data_window"]
    data_plot = data_item["data_plot"]
    data_options = data_item["data_options"]

    # check plot framework
    datachecker.check_choice("framework", framework, ["matplotlib", "pyvista"])

    # check title
    datachecker.check_string("title", title, can_be_empty=False)

    # check window data
    _check_data_window(data_window)

    # check options
    if framework == "pyvista":
        _check_data_options_pyvista(data_options)
    elif framework == "matplotlib":
        _check_data_options_matplotlib(data_options)
    else:
        raise ValueError("invalid plot framework")

    # check content
    if handler == "plotter":
        _check_data_plot_plotter(layout, data_plot)
    elif handler == "viewer":
        _check_data_plot_viewer(layout, data_plot)
    else:
        raise ValueError("invalid plot framework")


def check_data_plotter(data_plotter):
    """
    Check the type of the data defining several plots.
    This function is used for the plotter.
    """

    # check type
    datachecker.check_dict("data_plotter", data_plotter)

    # check items
    for data_plotter_tmp in data_plotter.values():
        _check_data_item(data_plotter_tmp, "plotter")


def check_data_viewer(data_viewer):
    """
    Check the type of the data defining several plots.
    This function is used for the viewer.
    """

    # check type
    datachecker.check_dict("data_viewer", data_viewer)

    # check items
    for data_viewer_tmp in data_viewer.values():
        _check_data_item(data_viewer_tmp, "viewer")



"""
Get the dense matrices used for the PEEC problem:
    - the integral of the Green functions
    - the cross-coupling functions between the faces
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import numpy.linalg as lna
from pypeec.lib_matrix import green_function


def _get_coupling(d, idx, method, dimension):
    """
    Compute a coupling function between two voxels for a specified coupling (direction of the faces).
    An analytical solution or a numerical approximation is used.
    """

    # extract the voxel data
    (dx, dy, dz) = d
    d = np.array(d, np.float_)

    # dimension permutation:
    #   - the 5D integral is solved for the xy faces
    #   - other faces are permuted to the xy faces
    if dimension == "xy":
        perm = [0, 1, 2]
    elif dimension == "xz":
        perm = [0, 2, 1]
    elif dimension == "yz":
        perm = [1, 2, 0]
    else:
        raise ValueError("invalid dimension")

    # make the permutation
    d_tmp = d[perm]
    idx_tmp = idx[:, perm]

    # shift for the integration of the last dimension
    idx_add = np.array([[0.0, 0.0, 0.5]])

    # compute the evaluation coordinate
    idx_1 = idx_tmp+idx_add
    idx_2 = idx_tmp-idx_add

    # get the partially integrated coefficients (5D integration)
    if method == "ana":
        G_1 = green_function.get_green_ana(d_tmp, idx_1, "5D")
        G_2 = green_function.get_green_ana(d_tmp, idx_2, "5D")
    elif method == "num":
        G_1 = green_function.get_green_num(d_tmp, idx_1, "5D")
        G_2 = green_function.get_green_num(d_tmp, idx_2, "5D")
    else:
        raise ValueError("invalid method")

    # get the 6D integral from the 5D integrals
    #   - the last dimension is the integral of the derivative
    #   - therefore, the integral is reduced to a subtraction
    G = G_2-G_1

    # scale the coefficients with the area of the faces
    if dimension == "xy":
        G = +G/(dy*dz*dx*dz)
    elif dimension == "xz":
        G = -G/(dy*dz*dx*dy)
    elif dimension == "yz":
        G = +G/(dx*dz*dx*dy)
    else:
        raise ValueError("invalid dimension")

    return G


def _get_voxel_indices(n):
    """
    Compute the indices of the complete voxel structure.
    Return the indices as a matrix.
    """

    # extract the voxel data
    (nx, ny, nz) = n

    # get the indices array
    idx_x = np.arange(nx, dtype=np.int_)
    idx_y = np.arange(ny, dtype=np.int_)
    idx_z = np.arange(nz, dtype=np.int_)
    [idx_x, idx_y, idx_z] = np.meshgrid(idx_x, idx_y, idx_z, indexing="ij")

    # flatten the indices into vectors
    idx_x = idx_x.flatten(order="F")
    idx_y = idx_y.flatten(order="F")
    idx_z = idx_z.flatten(order="F")

    # assemble the vectors in a matrix
    idx = np.stack((idx_x, idx_y, idx_z), axis=1)

    return idx


def _get_voxel_distances(d, idx):
    """
    Compute the normalized distance between the voxels and the reference voxel at the origin.
    """

    # scale the distances
    d_idx = d*idx

    # compute the distances
    d_cell = lna.norm(d_idx, axis=1)

    # normalize the distances
    n_cell = d_cell/max(d)

    return n_cell


def get_green_self(d):
    """
    Compute the self-coefficient for the Green functions.
    The self-coefficient is used for the preconditioner.
    """

    idx = np.array([[0, 0, 0]], dtype=np.int_)
    G_self = green_function.get_green_ana(d, idx, "6D")
    G_self = G_self[0]

    return G_self


def get_green_tensor(n, d, green_simplify):
    """
    Compute the Green functions for the complete voxel structure.
    For the self-coefficient and the close mutual coefficients, an analytical solution is used.
    For the remote mutual coefficients, an approximation is used.

    The voxel structure has the following size: (nx, ny, nz).
    The created tensor has the following dimension: (nx, ny, nz, 1).
    The self-coefficient is at the following location: (0, 0, 0, 0).
    All the elements are computed with respect to the first voxel.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    nv = nx*ny*nz

    # get the indices of the complete voxel structure (as a matrix)
    idx = _get_voxel_indices(n)

    # compute the normalized distance between the voxels and the reference voxel at the origin
    n_cell = _get_voxel_distances(d, idx)

    # check where the analytical solution should be used
    idx_ana = n_cell <= green_simplify
    idx_num = np.invert(idx_ana)

    # init the result vector
    G_mutual = np.empty(nv, dtype=np.float_)

    # analytical solution
    G_mutual[idx_ana] = green_function.get_green_ana(d, idx[idx_ana], "6D")

    # numerical solution
    G_mutual[idx_num] = green_function.get_green_num(d, idx[idx_num], "6D")

    # transform the vector into a tensor
    G_mutual = G_mutual.reshape((nx, ny, nz, 1), order="F")

    return G_mutual


def get_coupling_tensor(n, d, coupling_simplify, has_coupling):
    """
    Compute the coupling functions for the complete voxel structure.
    For the close coefficients, an analytical solution is used.
    For the remote coefficients, an approximation is used.

    The voxel structure has the following size: (nx, ny, nz).
    The created tensor has the following dimension: (nx, ny, nz, 3).
    All the elements are computed with respect to the first voxel.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    nv = nx*ny*nz

    # check if the tensor is required
    if not has_coupling:
        return None

    # get the indices of the complete voxel structure (as a matrix)
    idx = _get_voxel_indices(n)

    # compute the normalized distance between the voxels and the reference voxel at the origin
    n_cell = _get_voxel_distances(d, idx)

    # check where the analytical solution should be used
    idx_ana = n_cell <= coupling_simplify
    idx_num = np.invert(idx_ana)

    # init the result vector
    K_tsr = np.empty((nv, 3), dtype=np.float_)

    # analytical solution
    K_tsr[idx_ana, 0] = _get_coupling(d, idx[idx_ana], "ana", "yz")
    K_tsr[idx_ana, 1] = _get_coupling(d, idx[idx_ana], "ana", "xz")
    K_tsr[idx_ana, 2] = _get_coupling(d, idx[idx_ana], "ana", "xy")

    # numerical solution
    K_tsr[idx_num, 0] = _get_coupling(d, idx[idx_num], "num", "yz")
    K_tsr[idx_num, 1] = _get_coupling(d, idx[idx_num], "num", "xz")
    K_tsr[idx_num, 2] = _get_coupling(d, idx[idx_num], "num", "xy")

    # transform the vector into a tensor
    K_tsr = K_tsr.flatten(order="F")
    K_tsr = K_tsr.reshape((nx, ny, nz, 3), order="F")

    return K_tsr



"""
Main script for creating a 3D voxel structure.
Check the input data and mesh the structure.
The different parts of the code are timed.

Three different kind of geometry can be meshed:
    - stl: the voxel structure is generated from 3D STL files
    - png: the voxel structure is generated from stacked PNG images
    - shape: the voxel structure is generated from stacked 2D vector shapes
    - voxel: the voxel structure is given with the voxel indices

The mesher is implemented with PyVista, Pillow, Shapely, and Rasterio.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

from pypeec.lib_mesher import mesher_voxel
from pypeec.lib_mesher import mesher_shape
from pypeec.lib_mesher import mesher_png
from pypeec.lib_mesher import mesher_stl
from pypeec.lib_mesher import voxel_cloud
from pypeec.lib_mesher import voxel_conflict
from pypeec.lib_mesher import voxel_resample
from pypeec.lib_mesher import voxel_connection
from pypeec.lib_mesher import voxel_summary
from pypeec.lib_check import check_data_geometry
from pypeec.lib_check import check_data_options
from pypeec import log

# get a logger
LOGGER = log.get_logger("MESHER")


def _run_mesher(data_geometry):
    """
    Run the mesher
    """

    # extract the input data
    mesh_type = data_geometry["mesh_type"]
    data_voxelize = data_geometry["data_voxelize"]

    # voxelize the geometry
    if mesh_type == "voxel":
        (reference, data_internal) = _run_voxel(data_voxelize)
    elif mesh_type == "shape":
        (reference, data_internal) = _run_shape(data_voxelize)
    elif mesh_type == "png":
        (reference, data_internal) = _run_png(data_voxelize)
    elif mesh_type == "stl":
        (reference, data_internal) = _run_stl(data_voxelize)
    else:
        raise ValueError("invalid mesh type")

    return reference, data_internal


def _run_voxel(data_voxelize):
    """
    Generate a 3D voxel structure from indices.
    """

    # extract the data
    param = data_voxelize["param"]
    domain_index = data_voxelize["domain_index"]

    # process the indices arrays
    with log.BlockTimer(LOGGER, "mesher_voxel"):
        (n, d, c, domain_def, reference) = mesher_voxel.get_mesh(param, domain_index)

    # assemble the data
    data_internal = {
        "n": n, "d": d, "c": c,
        "domain_def": domain_def,
    }

    return reference, data_internal


def _run_shape(data_voxelize):
    """
    Generate a 3D voxel structure from 2D vector shapes.
    """

    # extract the data
    param = data_voxelize["param"]
    layer_stack = data_voxelize["layer_stack"]
    geometry_shape = data_voxelize["geometry_shape"]

    # process the shapes
    with log.BlockTimer(LOGGER, "mesher_shape"):
        (n, d, c, domain_def, reference) = mesher_shape.get_mesh(param, layer_stack, geometry_shape)

    # assemble the data
    data_internal = {
        "n": n, "d": d, "c": c,
        "domain_def": domain_def,
    }

    return reference, data_internal


def _run_png(data_voxelize):
    """
    Generate a 3D voxel structure from PNG images.
    """

    # extract the data
    param = data_voxelize["param"]
    domain_color = data_voxelize["domain_color"]
    layer_stack = data_voxelize["layer_stack"]

    # voxelize the PNG files
    with log.BlockTimer(LOGGER, "mesher_png"):
        (n, d, c, domain_def, reference) = mesher_png.get_mesh(param, domain_color, layer_stack)

    # assemble the data
    data_internal = {
        "n": n, "d": d, "c": c,
        "domain_def": domain_def,
    }

    return reference, data_internal


def _run_stl(data_voxelize):
    """
    Generate a 3D voxel structure from 3D STL files.
    """

    # extract the data
    param = data_voxelize["param"]
    domain_stl = data_voxelize["domain_stl"]

    # voxelize the STL files
    with log.BlockTimer(LOGGER, "mesher_stl"):
        (n, d, c, domain_def, reference) = mesher_stl.get_mesh(param, domain_stl)

    # assemble the data
    data_internal = {
        "n": n, "d": d, "c": c,
        "domain_def": domain_def,
    }

    return reference, data_internal


def _run_resample_graph(reference, data_internal, data_geometry, is_truncated):
    """
    Resampling of a 3D voxel structure (increases the number of voxels).
    """

    # extract the data
    resampling = data_geometry["resampling"]
    check_cloud = data_geometry["check_cloud"]
    check_conflict = data_geometry["check_conflict"]
    check_connection = data_geometry["check_connection"]
    domain_connection = data_geometry["domain_connection"]
    domain_conflict = data_geometry["domain_conflict"]
    pts_cloud = data_geometry["pts_cloud"]

    # extract the data
    n = data_internal["n"]
    d = data_internal["d"]
    c = data_internal["c"]
    domain_def = data_internal["domain_def"]

    with log.BlockTimer(LOGGER, "voxel_resample"):
        (n, d, c, s, domain_def) = voxel_resample.get_remesh(n, d, c, domain_def, resampling)

    if check_cloud:
        with log.BlockTimer(LOGGER, "voxel_cloud"):
            pts_cloud = voxel_cloud.get_cloud(n, d, c, domain_def, pts_cloud)

    if check_conflict:
        with log.BlockTimer(LOGGER, "voxel_conflict"):
            domain_def = voxel_conflict.get_conflict(domain_def, domain_conflict)

    if check_connection:
        with log.BlockTimer(LOGGER, "voxel_connection"):
            connection_def = voxel_connection.get_connection(n, domain_def, domain_connection)
    else:
        connection_def = []

    with log.BlockTimer(LOGGER, "voxel_summary"):
        voxel_status = voxel_summary.get_status(n, d, s, c, domain_def, connection_def)

    # assemble the data
    data_geom = {
        "n": n, "d": d, "s": s, "c": c,
        "voxel_status": voxel_status,
    }

    # if required, add the complete data
    if not is_truncated:
        data_add = {
            "domain_def": domain_def,
            "connection_def": connection_def,
            "pts_cloud": pts_cloud,
            "reference": reference,
        }
        data_geom = {**data_geom, **data_add}

    return data_geom


def _get_data(data_geom, timestamp, is_truncated):
    """
    Assemble the returned data.
    """

    # end message
    (seconds, duration, date) = log.get_duration(timestamp)

    # get status
    is_successful = True

    # extract the solution
    data_voxel = {
        "date": date,
        "duration": duration,
        "seconds": seconds,
        "is_truncated": is_truncated,
        "is_successful": is_successful,
        "data_geom": data_geom,
    }

    return data_voxel


def run(data_geometry, is_truncated=False):
    """
    Main script for meshing the geometry and generating a 3D voxel structure.
    Handle invalid data with exceptions.
    """

    # get timestamp
    timestamp = log.get_timer()

    # check the input data
    LOGGER.info("check the input data")
    check_data_geometry.check_data_geometry(data_geometry)
    check_data_options.check_data_options(is_truncated)

    # run the mesher
    (reference, data_internal) = _run_mesher(data_geometry)

    # resample and assemble
    data_geom = _run_resample_graph(reference, data_internal, data_geometry, is_truncated)

    # create output data
    data_voxel = _get_data(data_geom, timestamp, is_truncated)

    return data_voxel



"""
User script for visualizing a 3D voxel structure.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import sys
import os.path
from pypeec import main
import examples_config

# get config
PATH_ROOT = examples_config.PATH_ROOT
FOLDER_CONFIG = examples_config.FOLDER_CONFIG
FOLDER_EXAMPLE = examples_config.FOLDER_EXAMPLE


if __name__ == "__main__":
    # get the filenames
    file_voxel = os.path.join(PATH_ROOT, FOLDER_EXAMPLE, "voxel.gz")
    file_viewer = os.path.join(PATH_ROOT, FOLDER_CONFIG, "viewer.yaml")

    # plot tag (from viewer.yaml)
    tag_plot = ["domain", "connection"]

    # run viewer
    try:
        main.run_viewer_file(
            file_voxel, file_viewer,
            tag_plot=tag_plot,
            plot_mode="qt",
        )
    except Exception:
        sys.exit(1)
    else:
        sys.exit(0)



"""
Different functions for handling the problem values:
    - materials properties (electric and magnetic materials)
    - source values and internal impedance/admittance
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
from pypeec import log

# get a logger
LOGGER = log.get_logger("PROBLEM")


def get_material_vector(material_val, material_idx):
    """
    Get the material parameters for the different material types.
    """

    # array for the resistivities
    rho_vc = np.empty((0, 3), dtype=np.complex_)
    rho_vm = np.empty((0, 3), dtype=np.complex_)

    # populate the arrays
    for tag, material_idx_tmp in material_idx.items():
        # extract the data
        material_type = material_idx_tmp["material_type"]
        idx = material_idx_tmp["idx"]

        # vacuum permeability
        mu = 4*np.pi*1e-7

        # get the resistivities for the electric materials
        if material_type in ["electric", "electromagnetic"]:
            # assemble
            rho_re = material_val[tag]["rho_re"]
            rho_im = material_val[tag]["rho_im"]
            rho = rho_re+1j*rho_im

            # check size
            if len(rho) != len(idx):
                raise RuntimeError("invalid source")

            # append the resistivities
            rho_vc = np.concatenate((rho_vc, rho))

        # get the resistivities for the magnetic materials
        if material_type in ["magnetic", "electromagnetic"]:
            # assemble
            chi_re = material_val[tag]["chi_re"]
            chi_im = material_val[tag]["chi_im"]
            chi = chi_re-1j*chi_im
            rho = 1/(mu*chi)

            # check size
            if len(rho) != len(idx):
                raise RuntimeError("invalid material parameters")

            # append the resistivities
            rho_vm = np.concatenate((rho_vm, rho))

    return rho_vc, rho_vm


def get_source_all(source_val, source_pos, source_idx):
    """
    Merge the different source structures.
    """

    # init combined array
    source_all = {}

    # populate the arrays with the current sources
    for tag in source_idx:
        # extract the data
        source_type = source_idx[tag]["source_type"]
        var_type = source_idx[tag]["var_type"]
        idx = source_idx[tag]["idx"]

        # extract the data
        idx_src = source_pos[tag]["idx_src"]
        idx_vc = source_pos[tag]["idx_vc"]

        # get the values and admittances/impedances for the source
        if source_type == "current":
            # extract the data
            I_re = source_val[tag]["I_re"]
            I_im = source_val[tag]["I_im"]
            Y_re = source_val[tag]["Y_re"]
            Y_im = source_val[tag]["Y_im"]
            value = I_re+1j*I_im
            element = Y_re+1j*Y_im

            # scale the lumped parameters
            if var_type == "lumped":
                value_all = value/len(idx)
                element_all = element/len(idx)
            elif var_type == "distributed":
                value_all = value
                element_all = element
            else:
                raise ValueError("invalid variable type")
        elif source_type == "voltage":
            # extract the data
            V_re = source_val[tag]["V_re"]
            V_im = source_val[tag]["V_im"]
            Z_re = source_val[tag]["Z_re"]
            Z_im = source_val[tag]["Z_im"]
            value = V_re+1j*V_im
            element = Z_re+1j*Z_im

            # scale the lumped parameters
            if var_type == "lumped":
                value_all = value
                element_all = element*len(idx)
            elif var_type == "distributed":
                value_all = value
                element_all = element
            else:
                raise ValueError("invalid variable type")
        else:
            raise ValueError("invalid source type")

        # assign merged data
        source_all[tag] = {
            "source_type": source_type,
            "var_type": var_type,
            "idx": idx,
            "idx_src": idx_src,
            "idx_vc": idx_vc,
            "value": value_all,
            "element": element_all,
        }

    return source_all


def get_source_vector(source_all, source_type_ref):
    """
    Get the source parameters for a given source type.
    """

    # array for the source indices
    value_src = np.empty(0, dtype=np.complex_)
    element_src = np.empty(0, dtype=np.complex_)

    # populate the arrays with the current sources
    for tag, source_idx_tmp in source_all.items():
        # extract the data
        source_type = source_idx_tmp["source_type"]
        value = source_idx_tmp["value"]
        element = source_idx_tmp["element"]
        idx = source_idx_tmp["idx"]

        # check size
        if len(value) != len(idx):
            raise RuntimeError("invalid source parameters")
        if len(element) != len(idx):
            raise RuntimeError("invalid source parameters")

        # append the source values and admittances/impedances
        if source_type == source_type_ref:
            value_src = np.concatenate((value_src, value))
            element_src = np.concatenate((element_src, element))

    return value_src, element_src


def get_resistance_vector(n, d, A_net, idx_f, rho_v):
    """
    Extract the resistance vector of the system (diagonal of the resistance matrix).

    The problem contains n_f internal faces.
    The resistance vector has the following length: n_f.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    (dx, dy, dz) = d
    nv = nx*ny*nz

    # get the resistivity of the faces (average between voxels)
    rho = 0.5 * rho_v.transpose()*np.abs(A_net)

    # get the direction of the faces (x, y, z)
    idx_fx = np.in1d(idx_f, np.arange(0*nv, 1*nv, dtype=np.int_))
    idx_fy = np.in1d(idx_f, np.arange(1*nv, 2*nv, dtype=np.int_))
    idx_fz = np.in1d(idx_f, np.arange(2*nv, 3*nv, dtype=np.int_))

    # resistance vector (different directions)
    R = np.zeros(len(idx_f), dtype=np.complex_)
    R[idx_fx] = (dx/(dy*dz))*rho[0, idx_fx]
    R[idx_fy] = (dy/(dx*dz))*rho[1, idx_fy]
    R[idx_fz] = (dz/(dx*dy))*rho[2, idx_fz]

    return R



"""
Different functions for dealing the voxel data.
Compute the voxel coordinates and the incidence matrix.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import scipy.sparse as sps


def get_voxel_coordinate(n, d, c):
    """
    Get the coordinate of the different voxels.
    The center voxel center is at the specified origin coordinate.

    The voxel structure has the following size: (nx, ny, nz).
    The array has the following dimension: (nx*ny*nz, 3).
    """

    # cast to array
    c = np.array(c, dtype=np.float_)
    d = np.array(d, dtype=np.float_)
    n = np.array(n, dtype=np.int_)

    # all the indices
    idx_linear = np.arange(0, np.prod(n), dtype=np.int_)

    # convert linear indices into tensor indices
    (idx_x, idx_y, idx_z) = np.unravel_index(idx_linear, n, order="F")

    # assemble the coordinate array
    idx_vox = np.stack((idx_x, idx_y, idx_z), axis=1)

    # origin coordinate
    o = c-(n*d)/2

    # assemble the coordinate array
    pts_vox = o+d/2+d*idx_vox

    return pts_vox


def get_incidence_matrix(n):
    """
    Get the incidence matrix of the voxel structure.
    This matrix describes the relation between the voxels and the faces.

    The voxel structure has the following size: (nx, ny, nz).
    The matrix has the following dimension: (nx*ny*nz, 3*nx*ny*nz).
    The columns represent the face with the following order: x, y, and, z.
    The rows represent the voxels.
    """

    # extract the voxel data
    (nx, ny, nz) = n
    nv = nx*ny*nz

    # voxel index array
    x = np.arange(nx, dtype=np.int_)
    y = np.arange(ny, dtype=np.int_)
    z = np.arange(nz, dtype=np.int_)
    (idx_x, idx_y, idx_z) = np.meshgrid(x, y, z, indexing="ij")

    # voxel index number
    idx = idx_x+idx_y*nx+idx_z*nx*ny

    # create the sparse matrix
    A_vox = sps.csc_matrix((nv, 3*nv), dtype=np.int_)

    # assign the diagonal, each voxel is connected to three faces with positive indices
    data = np.ones(nv)
    idx_row_col = np.arange(nv, dtype=np.int_)
    A_vox += sps.csc_matrix((data, (idx_row_col, 0*nv+idx_row_col)), shape=(nv, 3*nv), dtype=np.int_)
    A_vox += sps.csc_matrix((data, (idx_row_col, 1*nv+idx_row_col)), shape=(nv, 3*nv), dtype=np.int_)
    A_vox += sps.csc_matrix((data, (idx_row_col, 2*nv+idx_row_col)), shape=(nv, 3*nv), dtype=np.int_)

    # faces along x direction (faces with negative indices)
    idx_col = idx[:-1, :, :].flatten()
    idx_row = idx[+1:, :, :].flatten()
    data = -np.ones((nx-1)*ny*nz, dtype=np.int_)
    A_vox += sps.csc_matrix((data, (idx_row, 0*nv+idx_col)), shape=(nv, 3*nv), dtype=np.int_)

    # faces along y direction (faces with negative indices)
    idx_col = idx[:, :-1, :].flatten()
    idx_row = idx[:, +1:, :].flatten()
    data = -np.ones(nx*(ny-1)*nz, dtype=np.int_)
    A_vox += sps.csc_matrix((data, (idx_row, 1*nv+idx_col)), shape=(nv, 3*nv), dtype=np.int_)

    # faces along z direction (faces with negative indices)
    idx_col = idx[:, :, :-1].flatten()
    idx_row = idx[:, :, +1:].flatten()
    data = -np.ones(nx*ny*(nz-1), dtype=np.int_)
    A_vox += sps.csc_matrix((data, (idx_row, 2*nv+idx_col)), shape=(nv, 3*nv), dtype=np.int_)

    return A_vox



"""
Different functions for computing Green functions with voxels.
Analytical solutions and numerical approximations are used.

The two following integrals are available:
    - integrate all the dimensions (6D integrals)
    - integrate the dimensions except the last one (5D integrals)

Exact Inductance Equations for Rectangular Conductors With Applications to More Complicated Geometries
C. Hoer and C. Love, 1965

Exact Closed Form Formula for Self Inductance of Conductor of Rectangular Cross Section
Z. Piatek and B. Baron, 2021
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import numpy as np
import numpy.linalg as lna
from pypeec import log

# get a logger
LOGGER = log.get_logger("GREEN")


def _get_safe_inv(x):
    """
    Compute the inverse. Set to zero if not finite.
    """

    y = 1/x
    y[np.isnan(y)] = 0.0
    y[np.isinf(y)] = 0.0

    return y


def _get_safe_log(x):
    """
    Compute the log. Set to zero if not finite.
    """

    y = np.log(x)
    y[np.isnan(y)] = 0.0
    y[np.isinf(y)] = 0.0

    return y


def _get_safe_arctan(x):
    """
    Compute the arctan. Set to zero if not finite.
    """

    y = np.arctan(x)
    y[np.isnan(y)] = 0.0
    y[np.isinf(y)] = 0.0

    return y


def _get_green_6D_fct(x, y, z):
    """
    Compute the analytical Green function between two voxels.
    All the dimensions are integrated (6D integrals).
    The analytical solution for points is computed for a given distances.
    """

    # precompute terms
    with np.errstate(all="ignore"):
        nrm = np.sqrt(x**2+y**2+z**2)
        atanx = _get_safe_arctan((y*z)/(x*nrm))
        atany = _get_safe_arctan((x*z)/(y*nrm))
        atanz = _get_safe_arctan((x*y)/(z*nrm))
        logx = _get_safe_log(x+nrm)
        logy = _get_safe_log(y+nrm)
        logz = _get_safe_log(z+nrm)

    # compute function
    val = 1.0*(
            + (x**4*nrm)/60 +
            + (y**4*nrm)/60 +
            + (z**4*nrm)/60 +
            - (x*y**4*logx)/24 +
            - (x*z**4*logx)/24 +
            - (x**4*y*logy)/24 +
            - (y*z**4*logy)/24 +
            - (x**4*z*logz)/24 +
            - (y**4*z*logz)/24 +
            - (x**2*y**2*nrm)/20 +
            - (x**2*z**2*nrm)/20 +
            - (y**2*z**2*nrm)/20 +
            + (x*y**2*z**2*logx)/4 +
            + (x**2*y*z**2*logy)/4 +
            + (x**2*y**2*z*logz)/4 +
            - (x*y*z**3*atanz)/6 +
            - (x*y**3*z*atany)/6 +
            - (x**3*y*z*atanx)/6
    )

    return val


def _get_green_5D_fct(x, y, z):
    """
    Compute the analytical Green function between two voxels.
    All the dimensions except the last one are integrated (5D integrals).
    The analytical solution for points is computed for a given distances.
    """

    # precompute terms
    with np.errstate(all="ignore"):
        nrm = np.sqrt(x**2+y**2+z**2)
        inv = _get_safe_inv(nrm)
        invx = _get_safe_inv(x+nrm)
        invy = _get_safe_inv(y+nrm)
        invxz = _get_safe_inv(x**2+z**2)
        invyz = _get_safe_inv(y**2+z**2)
        invxyz = _get_safe_inv(x**2*y**2+x**2*z**2+y**2*z**2+z**4)
        logx = _get_safe_log(x+nrm)
        logy = _get_safe_log(y+nrm)
        logz = _get_safe_log(z+nrm)
        atany = _get_safe_arctan((x*z)/(y*nrm))
        atanx = _get_safe_arctan((y*z)/(x*nrm))
        atanz = _get_safe_arctan((x*y)/(z*nrm))

    # compute function
    val = 1.0*(
            +(z**3*nrm)/15 +
            + (z**5*inv)/60 +
            - (x**4*logz)/24 +
            - (y**4*logz)/24 +
            - (x*z**3*logx)/6 +
            - (y*z**3*logy)/6 +
            + (x**2*y**2*logz)/4 +
            + (x*y**2*z*logx)/2 +
            + (x**2*y*z*logy)/2 +
            - (x*y**3*atany)/6 +
            - (x**3*y*atanx)/6 +
            - (x*y*z**2*atanz)/2 +
            - (x**2*z**3*inv)/20 +
            - (y**2*z**3*inv)/20 +
            - (x**2*z*nrm)/10 +
            - (y**2*z*nrm)/10 +
            - (x**4*z*inv)/40 +
            - (y**4*z*inv)/40 +
            + (x**2*y**2*z*inv)/5 +
            - (x*z**5*inv*invx)/24 +
            - (y*z**5*inv*invy)/24 +
            - (x**4*y**2*z*inv*invxz)/6 +
            - (x**2*y**4*z*inv*invyz)/6 +
            + (x*y**2*z**3*inv*invx)/4 +
            + (x**2*y*z**3*inv*invy)/4 +
            - (x*y**4*z*inv*invx)/24 +
            - (x**4*y*z*inv*invy)/24 +
            + (x**2*y**2*z**5*inv*invxyz)/3 +
            + (x**2*y**4*z**3*inv*invxyz)/6 +
            + (x**4*y**2*z**3*inv*invxyz)/6
    )

    return val


def _get_green_preproc(int_type):
    """
    Compute the integral boundaries for the Green functions.
    Offset vectors, which are used to compute the distances between points, are generated.
    A sign vector, which is used to sum (integrate) the points, is generated.

    For the 6D integral, 64 boundaries points are generated.
    For the 5D integral, 32 boundaries points are generated.
    """

    # get the offset vector
    if int_type == "6D":
        offset_xy = np.array([-1.0, 0.0, +1.0, 0.0], dtype=np.float_)
        offset_z = np.array([-1.0, 0.0, +1.0, 0.0], dtype=np.float_)
        idx_xy = np.arange(1, 5, dtype=np.int_)
        idx_z = np.arange(1, 5, dtype=np.int_)
    elif int_type == "5D":
        offset_xy = np.array([-1.0, 0.0, +1.0, 0.0], dtype=np.float_)
        offset_z = np.array([+0.5, -0.5], dtype=np.float_)
        idx_xy = np.arange(1, 5, dtype=np.int_)
        idx_z = np.arange(1, 3, dtype=np.int_)
    else:
        raise ValueError("invalid integral type")

    (offset_x, offset_y, offset_z) = np.meshgrid(offset_xy, offset_xy, offset_z)
    offset_x = offset_x.flatten()
    offset_y = offset_y.flatten()
    offset_z = offset_z.flatten()

    # get the sign indices
    (idx_x, idx_y, idx_z) = np.meshgrid(idx_xy, idx_xy, idx_z)
    idx_x = idx_x.flatten()
    idx_y = idx_y.flatten()
    idx_z = idx_z.flatten()

    # get the sign vector
    sign = (-1)**(idx_x+idx_y+idx_z+1)

    return offset_x, offset_y, offset_z, sign


def get_green_ana(d, idx, int_type):
    """
    Compute a Green function between two voxels.
    An analytical solution is used.
    The 5D or 6D integrals can be computed.
    """

    # check if empty
    if len(idx) == 0:
        return np.empty(0, dtype=np.float_)

    # display
    LOGGER.debug("analytical solution: %s / %d" % (int_type, len(idx)))

    # extract the voxel data
    (dx, dy, dz) = d

    # extract the voxel distance
    idx_x = idx[:, [0]]
    idx_y = idx[:, [1]]
    idx_z = idx[:, [2]]

    # get the offset and sign vectors
    (offset_x, offset_y, offset_z, sign) = _get_green_preproc(int_type)

    # position vector
    x_vec = dx*(idx_x+offset_x)
    y_vec = dy*(idx_y+offset_y)
    z_vec = dz*(idx_z+offset_z)

    # compute the values
    if int_type == "6D":
        val = _get_green_6D_fct(x_vec, y_vec, z_vec)
    elif int_type == "5D":
        val = _get_green_5D_fct(x_vec, y_vec, z_vec)
    else:
        raise ValueError("invalid integral type")

    # sum the value of all the points
    G = np.sum(sign*val, axis=1)

    # add scaling
    G = G/(4*np.pi)

    return G


def get_green_num(d, idx, int_type):
    """
    Compute a Green function between two voxels.
    The 5D or 6D integrals can be computed.
    A numerical approximation is used.
    Only valid if the distance between the two voxels is large.
    """

    # check if empty
    if len(idx) == 0:
        return np.empty(0, dtype=np.float_)

    # display
    LOGGER.debug("numerical approximation: %s / %d" % (int_type, len(idx)))

    # extract the voxel data
    (dx, dy, dz) = d

    # compute the volume and the distance
    vol = np.prod(d)

    # compute the physical distance
    dis = lna.norm(np.multiply(d, idx), axis=1)

    # compute the approximation
    if int_type == "6D":
        G = (vol*vol)/(4*np.pi*dis)
    elif int_type == "5D":
        G = (dx*dy*vol)/(4*np.pi*dis)
    else:
        raise ValueError("invalid integral type")

    return G



"""
User script for meshing a voxel structure.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import sys
import os.path
from pypeec import main
import examples_config

# get config
PATH_ROOT = examples_config.PATH_ROOT
FOLDER_CONFIG = examples_config.FOLDER_CONFIG
FOLDER_EXAMPLE = examples_config.FOLDER_EXAMPLE


if __name__ == "__main__":
    # get the filenames
    file_geometry = os.path.join(PATH_ROOT, FOLDER_EXAMPLE, "geometry.yaml")
    file_voxel = os.path.join(PATH_ROOT, FOLDER_EXAMPLE, "voxel.gz")

    # run
    try:
        main.run_mesher_file(file_geometry, file_voxel)
    except Exception:
        sys.exit(1)
    else:
        sys.exit(0)



"""
Module for running the PyPEEC workflow.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os.path
import tempfile
import logging
import warnings
from pypeec import main
from pypeec import script
from pypeec import io

# crash on warnings (except for deprecation warnings)
warnings.filterwarnings("error")
warnings.filterwarnings("ignore", category=DeprecationWarning)

# disable logging to prevent clutter during test evaluation
logging.disable(logging.INFO)

# get the path the folder
PATH_ROOT = os.path.dirname(__file__)

# construct the config folder path
FOLDER_EXAMPLES = os.path.join(PATH_ROOT, "..", "..", "examples")


def _create_temp_file():
    """
    Get a temporary file.
    """

    (_, filename) = tempfile.mkstemp(suffix=".gz")

    return filename


def _delete_temp_file(filename):
    """
    Delete a temporary file.
    """

    try:
        os.remove(filename)
    except OSError:
        pass


def _run_script(argv):
    """
    Run the command line script with arguments.
    """

    status = script.run_arguments(argv)
    if status != 0:
        raise RuntimeError("invalid return code for the script")


def run_workflow(name, use_script):
    """
    Run the complete workflow:
        - run the mesher
        - run the viewer
        - run the solver
        - run the plotter

    The intermediate file are stored with temporary files.
    The temporary files are deleted at the end of the function.

    The workflow can be run with two modes:
        - with the command line script (pypeec.script)
        - with the API (pypeec.main)
    """

    # get input file name
    file_geometry = os.path.join(FOLDER_EXAMPLES, name, "geometry.yaml")
    file_problem = os.path.join(FOLDER_EXAMPLES, name, "problem.yaml")

    # get config file name
    file_plotter = os.path.join(FOLDER_EXAMPLES, "config", "plotter.yaml")
    file_viewer = os.path.join(FOLDER_EXAMPLES, "config", "viewer.yaml")
    file_tolerance = os.path.join(FOLDER_EXAMPLES, "config", "tolerance.yaml")

    # get the temporary files
    file_voxel = _create_temp_file()
    file_solution = _create_temp_file()

    # run the code
    try:
        # run the workflow
        if use_script:
            # get the arguments
            argv_me = ["me", "-ge", file_geometry, "-vo", file_voxel]
            argv_vi = ["vi", "-vo", file_voxel,  "-vi", file_viewer, "-pm", "none"]
            argv_so = ["so", "-vo", file_voxel,  "-pr", file_problem, "-to", file_tolerance, "-so", file_solution]
            argv_pl = ["pl", "-so", file_solution,  "-pl", file_plotter, "-pm", "none"]

            # run the scripts
            _run_script(argv_me)
            _run_script(argv_vi)
            _run_script(argv_so)
            _run_script(argv_pl)
        else:
            main.run_mesher_file(file_geometry, file_voxel)
            main.run_viewer_file(file_voxel, file_viewer, plot_mode="none")
            main.run_solver_file(file_voxel, file_problem, file_tolerance, file_solution)
            main.run_plotter_file(file_solution, file_plotter, plot_mode="none")

        # load the files
        data_voxel = io.load_data(file_voxel)
        data_solution = io.load_data(file_solution)
    finally:
        # close the temporary files
        _delete_temp_file(file_voxel)
        _delete_temp_file(file_solution)

    return data_voxel, data_solution



"""
Module for serialization and deserialization.
    - load JSON/YAML input files
    - load and write JSON/Pickle data files

For YAML files, the following custom extensions are used:
    - "!path" - parse relative paths (with respect to the main file path)
    - "!include" - include other YAML files (recursion possible)
    - "!env" - include YAML string from environment variables
    - "!merge_dict" - merge a list of dicts
    - "!merge_list" - merge a list of lists

For JSON files, the following custom extensions are used:
    - "__complex__" - allows the serialization of complex numbers
    - "__numpy__" - allows the serialization of NumPy arrays

The JSON/YAML files with the custom extensions are still valid JSON/YAML files.

The JSON files can be serialized/deserialized as/from:
    - text files
    - gzip files

Warning:
    - Pickling data is not secure.
    - Only load pickle files that you trust.
"""

__author__ = "Thomas Guillod"
__copyright__ = "Thomas Guillod - Dartmouth College"
__license__ = "Mozilla Public License Version 2.0"

import os.path
import json
import pickle
import gzip
import yaml
import numpy as np


class _MergeData:
    """
    This Python class is used to merge YAML data.
        - a custom merge command is used with a list of arguments
        - the arguments (lists or dicts) are merged together
        - the merge is performed recursively

    The merge objects are created during the YAML parsing.
    The merge objects are replaced by the merged data after the parsing.
    """

    def __init__(self, data_list, data_type):
        """
        Constructor.
        Assign the list of data to be merged and the data type.
        """

        if type(data_list) is not list:
            raise yaml.YAMLError("arguments of the merge_dict / merge_list should be a list")

        self.data_list = data_list
        self.data_type = data_type

    @staticmethod
    def merge(data):
        """
        Walk through the data recursively and merge it.
        Find the merge objects and replace them with merged data.
        """

        if type(data) is dict:
            for tag, val in data.items():
                data[tag] = _MergeData.merge(val)
        elif type(data) is list:
            for idx, val in enumerate(data):
                data[idx] = _MergeData.merge(val)
        elif type(data) is _MergeData:
            data = data.extract()
        else:
            pass

        return data

    def extract(self):
        """
        Merge a list of dicts or a list of lists.
        The merge is performed recursively.
        """

        if self.data_type == "dict":
            res = {}
            for data in self.data_list:
                data = _MergeData.merge(data)
                if type(data) is not dict:
                    raise yaml.YAMLError("merge_dict cannot only merge dictionaries")
                res.update(data)
        elif self.data_type == "list":
            res = []
            for data in self.data_list:
                data = _MergeData.merge(data)
                if type(data) is not list:
                    raise yaml.YAMLError("merge_list cannot only merge lists")
                res += data
        else:
            raise yaml.YAMLError("invalid merge type")

        return res


class _YamlLoader(yaml.Loader):
    """
    This Python class offers extension to the YAML format.
        - include YAML file in YAML file
        - include relative filesystem paths
        - merge list of dicts
        - merge list of lists
    """

    def __init__(self, stream):
        """
        Constructor.
        Custom YAML loader subclassing the default loader.
        """

        # get the path of the YAML file for relative paths
        self.path_root = os.path.abspath(stream.name)
        self.path_root = os.path.dirname(self.path_root)

        # flag indicating if any merge commands are used
        self.has_merge = False

        # call the constructor of the parent
        super().__init__(stream)

        # handling of YAML files inclusion
        def fct_handle_include(self, node):
            res = _YamlLoader._yaml_handling(self, node, self._extract_yaml)
            return res

        # handling of relative paths
        def fct_handle_path(self, node):
            res = _YamlLoader._yaml_handling(self, node, self._extract_path)
            return res

        # handling of environment variables inclusion
        def fct_handle_env(self, node):
            res = _YamlLoader._yaml_handling(self, node, self._extract_env)
            return res

        # handling merge of a list of dicts
        def fct_handle_merge_dict(self, node):
            self.has_merge = True
            res = _MergeData(self.construct_sequence(node), "dict")
            return res

        # handling merge of a list of lists
        def fct_handle_merge_list(self, node):
            self.has_merge = True
            res = _MergeData(self.construct_sequence(node), "list")
            return res

        # add the extension to the YAML format
        _YamlLoader.add_constructor("!include", fct_handle_include)
        _YamlLoader.add_constructor("!path", fct_handle_path)
        _YamlLoader.add_constructor("!env", fct_handle_env)
        _YamlLoader.add_constructor("!merge_dict", fct_handle_merge_dict)
        _YamlLoader.add_constructor("!merge_list", fct_handle_merge_list)

    def _yaml_handling(self, node, fct):
        """
        Apply a function to a YAML node for list, dict, scalar.
        """

        if isinstance(node, yaml.ScalarNode):
            return fct(self.construct_scalar(node))
        elif isinstance(node, yaml.SequenceNode):
            result = []
            for arg in self.construct_sequence(node):
                result.append(fct(arg))
            return result
        elif isinstance(node, yaml.MappingNode):
            result = {}
            for tag, arg in self.construct_mapping(node).items():
                result[tag] = fct(arg)
            return result
        else:
            raise yaml.YAMLError("invalid YAML node type")

    def _extract_path(self, filename):
        """
        Find the path with respect to the YAML file path.
        """

        # check type
        if type(filename) is not str:
            raise yaml.YAMLError("path command arguments should be strings")

        # construct relative path
        filepath = os.path.join(self.path_root, filename)

        return filepath

    def _extract_yaml(self, filename):
        """
        Load an included YAML file.
        """

        # check type
        if type(filename) is not str:
            raise yaml.YAMLError("include command arguments should be strings")

        # construct relative path
        filepath = os.path.join(self.path_root, filename)

        # load YAML file
        with open(filepath, "r") as fid:
            data = _parse_yaml(fid)

        return data

    def _extract_env(self, name):
        """
        Load a YAML file from an environment variable.
        """

        # check type
        if type(name) is not str:
            raise yaml.YAMLError("env command arguments should be strings")

        # get and check the variable
        value = os.getenv(name)
        if value is None:
            raise yaml.YAMLError("env variable is not existing: %s" % name)

        # load YAML string
        data = yaml.safe_load(value)

        return data


class _JsonNumPyEncoder(json.JSONEncoder):
    """
    This Python class offers extension to the JSON format (encoder).
        - encode NumPy scalar types
        - encode NumPy array types
    """

    def __init__(self, **kwargs):
        """
        Constructor
        """

        super().__init__(**kwargs)

    def default(self, obj):
        """
        Function encoding NumPy types as dictionaries.
        """

        # encode complex number and numpy array
        if np.isscalar(obj) and np.iscomplexobj(obj):
            return {
                "__complex__": None,
                "real": obj.real,
                "imag": obj.imag,
            }
        elif isinstance(obj, np.ndarray):
            # handle numpy array
            if np.issubdtype(obj.dtype, np.complex_):
                return {
                    "__numpy__": None,
                    "dtype": "complex",
                    "shape": obj.shape,
                    "data": {
                        "real": obj.real.flatten().tolist(),
                        "imag": obj.imag.flatten().tolist(),
                    },
                }
            elif np.issubdtype(obj.dtype, np.floating):
                return {
                    "__numpy__": None,
                    "dtype": "float",
                    "shape": obj.shape,
                    "data": obj.flatten().tolist(),
                }
            elif np.issubdtype(obj.dtype, np.integer):
                return {
                    "__numpy__": None,
                    "dtype": "int",
                    "shape": obj.shape,
                    "data": obj.flatten().tolist(),
                }
            elif np.issubdtype(obj.dtype, bool):
                return {
                    "__numpy__": None,
                    "dtype": "bool",
                    "shape": obj.shape,
                    "data": obj.flatten().tolist(),
                }
            else:
                TypeError("invalid numpy array for serialization")
        else:
            # if not numpy, default to the base encoder
            return json.JSONEncoder.default(self, obj)


class _JsonNumPyDecoder(json.JSONDecoder):
    """
    This Python class offers extension to the JSON format (decoder).
        - decode NumPy scalar types
        - decode NumPy array types
    """

    def __init__(self, **kwargs):
        """
        Constructor
        """

        kwargs.setdefault("object_hook", self.parse)
        super().__init__(**kwargs)

    def parse(self, obj):
        """
        Function decoding NumPy types from dictionaries.
        """

        # if not dict, do nothing
        if not isinstance(obj, dict):
            return obj

        # parse the extensions
        if "__complex__" in obj:
            # handling complex scalar
            real = obj["real"]
            imag = obj["imag"]
            return complex(real, imag)
        elif "__numpy__" in obj:
            # handle numpy array
            dtype = obj["dtype"]
            shape = obj["shape"]
            data = obj["data"]

            # parse the type
            if dtype == "complex":
                real = np.array(data["real"], dtype=np.complex_).reshape(shape)
                imag = np.array(data["imag"], dtype=np.complex_).reshape(shape)
                return real+1j*imag
            elif dtype == "float":
                return np.array(data, dtype=np.float_).reshape(shape)
            elif dtype == "int":
                return np.array(data, dtype=np.int_).reshape(shape)
            elif dtype == "bool":
                return np.array(data, dtype=bool).reshape(shape)
        else:
            return obj


def _parse_yaml(stream):
    """
    Load a YAML stream (with custom extensions).
    If required, merge the data (custom merge commands).
    """

    loader = _YamlLoader(stream)
    try:
        data = loader.get_single_data()
        if loader.has_merge:
            data = _MergeData.merge(data)
    finally:
        loader.dispose()

    return data


def _load_yaml(filename):
    """
    Load a YAML file (with custom extensions).
    """

    try:
        with open(filename, "r") as fid:
            data = _parse_yaml(fid)
    except yaml.YAMLError as ex:
        raise RuntimeError("invalid YAML file: %s\n%s" % (filename, str(ex)))
    except OSError:
        raise OSError("cannot open the file: %s" % filename)

    return data


def _load_json(filename, is_gzip):
    """
    Load a JSON file (with extensions).
    The JSON file can be a text file or a gzip file.
    """

    try:
        if is_gzip:
            with gzip.open(filename, "rt", encoding="utf-8") as fid:
                data = json.load(fid, cls=_JsonNumPyDecoder)
        else:
            with open(filename, "r") as fid:
                data = json.load(fid, cls=_JsonNumPyDecoder)
    except (json.JSONDecodeError, TypeError, ValueError) as ex:
        raise RuntimeError("invalid JSON file: %s\n%s" % (filename, str(ex)))
    except OSError:
        raise OSError("cannot open the file: %s" % filename)

    return data


def _write_json(filename, data, is_gzip):
    """
    Write a JSON file (with extensions).
    The JSON file can be a text file or a gzip file.
    """

    try:
        if is_gzip:
            with gzip.open(filename, "wt", encoding="utf-8") as fid:
                json.dump(data, fid, cls=_JsonNumPyEncoder)
        else:
            with open(filename, "w") as fid:
                json.dump(data, fid, indent=4, cls=_JsonNumPyEncoder)
    except (json.JSONDecodeError, TypeError, ValueError) as ex:
        raise RuntimeError("invalid JSON file: %s\n%s" % (filename, str(ex)))
    except OSError:
        raise OSError("cannot write the file: %s" % filename)

    return data


def _load_pickle(filename):
    """
    Load a pickle file.
    """

    # load the Pickle file
    try:
        with open(filename, "rb") as fid:
            data = pickle.load(fid)
    except pickle.UnpicklingError:
        raise RuntimeError("invalid Pickle file: %s" % filename)
    except EOFError:
        raise EOFError("file not found: %s" % filename)
    except OSError:
        raise OSError("invalid Pickle file: %s" % filename)

    return data


def _write_pickle(filename, data):
    """
    Write a pickle file.
    """

    # save the Pickle file
    try:
        with open(filename, "wb") as fid:
            pickle.dump(data, fid)
    except pickle.PicklingError:
        raise RuntimeError("invalid data for Pickle: %s" % filename)
    except OSError:
        raise OSError("cannot write the file: %s" % filename)


def load_input(filename):
    """
    Load an input file (JSON or YAML).

    Parameters
    ----------
    filename : string)
        Name and path of the file to be loaded.
        The file type is determined by the extension.
        For YAML files, the extension should be "yaml" or "yml".
        For JSON files, the extension should be "json" or "js".
        For GZIP/JSON files, the extension should be "gzip" or "gz".

    Returns
    -------
    data : data
        Python data contained in the file content
    """

    (name, ext) = os.path.splitext(filename)
    if ext in [".json", ".js"]:
        data = _load_json(filename, False)
    elif ext in [".gz", ".gzip"]:
        data = _load_json(filename, True)
    elif ext in [".yaml", ".yml"]:
        data = _load_yaml(filename)
    else:
        raise ValueError("invalid file extension: %s" % filename)

    return data


def load_data(filename):
    """
    Load a data file (JSON or Pickle).

    Parameters
    ----------
    filename : string)
        Name and path of the file to be loaded.
        The file type is determined by the extension.
        For JSON files, the extension should be "json" or "js".
        For GZIP/JSON files, the extension should be "gzip" or "gz".
        For Pickle files, the extension should be "pck".

    Returns
    -------
    data : data
        Python data contained in the file content
    """

    (name, ext) = os.path.splitext(filename)
    if ext in [".json", ".js"]:
        data = _load_json(filename, False)
    elif ext in [".gz", ".gzip"]:
        data = _load_json(filename, True)
    elif ext in [".pck", ".pkl"]:
        data = _load_pickle(filename)
    else:
        raise ValueError("invalid file extension: %s" % filename)

    return data


def write_data(filename, data):
    """
    Write a data file (JSON or Pickle).

    Parameters
    ----------
    filename : string)
        Name and path of the file to be created.
        The file type is determined by the extension.
        For JSON files, the extension should be "json" or "js".
        For GZIP/JSON files, the extension should be "gzip" or "gz".
        For Pickle files, the extension should be "pck".
    data : data
        Python data to be saved.
    """

    (name, ext) = os.path.splitext(filename)
    if ext in [".json", ".js"]:
        _write_json(filename, data, False)
    elif ext in [".gz", ".gzip"]:
        _write_json(filename, data, True)
    elif ext in [".pck", ".pkl"]:
        _write_pickle(filename, data)
    else:
        raise ValueError("invalid file extension: %s" % filename)
